package org.qtunes.auth;

import org.qtunes.core.*;
import java.util.*;
import java.net.*;

/**
 * The Auth class handles creation of Sessions and determining what rights
 * those Sessions have. Each QTunes object has only one Auth Service.
 * Different Services may register an {@link Authorizer}
 * with this object: when a Session is created by a call to {@link #login},
 * each Authorizer for the matching Service is queried to see what "allowances"
 * it gives it.
 */
public interface Auth extends Service {

    /**
     * Get the Session with the specified ID, or null if it doesn't exist
     */
    public Session getSession(int id);

    /**
     * Create a new Session
     * @param properties a Map of properties which are part of the session - the exact content
     * depends on the service (for example, a web service might include the Remote IP address
     * of the connection). These properties are tested by the various {@link Authorizer Authorizers}
     * registered with the system to determine the set of allowances. At a minimum the
     * Map must contain a "session" key with the value as either the name of the Service
     * or the Service itself.
     */
    public Session login(Map properties);

    /**
     * Expire the specified Session
     * @param session the session to expire
     */
    public void logout(Session session);

    /**
     * Add an Authorizer to the Auth object. Any session creation requests
     * will check with all the authorizers to see what rights that Session
     * has.
     */
    public void addAuthorizer(Authorizer auth);

    /**
     * Remove the previously added Authorizer from the Auth object
     */
    public void removeAuthorizer(Authorizer auth);

    /**
     * Expire all Sessions created by the specified Service
     */
    public void logoutSessions(Service service);

    /**
     * Return the set of allowances for the specified Session
     */
    public Collection<String> getAllowances(Session session);

}
