package org.qtunes.auth;

import org.qtunes.core.*;
import org.qtunes.core.util.AbstractMatcher;
import java.net.*;
import java.math.BigInteger;

/**
 * A {@link AbstractMatcher} which can be used to match a Session object.
 * Tests include
 * <table class="def">
 * <tr><th>service</th><td>the name of the service that created this Session</td></tr>
 * <tr><th>ttl</th><td>the number of seconds until this Service expires</td></tr>
 * <tr><th>age</td><td>the number of seconds since this Service was created</td></tr>
 * <tr><th>idle</th><td>the number of seconds this Service has been idle</td></tr>
 * <tr><th>traffic</th><td>the number of bytes this Session has transmitted</td></tr>
 * <tr><th>tracksStarted</th><td>the number of tracks this Session has started</td></tr>
 * <tr><th>tracksCompleted</th><td>the number of tracks this Session has completed</td></tr>
 * <tr><th>ip</th><td>if appropriate, the IP address of the session. This can be tested as ip="192.168.1.2", ip="192.168.0.0/16", ip="*.local" or a regular expression match</td></tr>
 * </table>
 * or any property specified in the Session properties. So as an example,
 * <code>service="AppleServer" and (remote="Mike's Remote" or ip="10.0.0.16")</code> could
 * be used to scrobble tracks played by a certain remote or streamed to a certain host.
 */
public class SessionMatcher extends AbstractMatcher<Session> {

    public SessionMatcher(String filter) {
        super(filter);
    }

    public Object getProperty(Session session, String property) {
        if (property.equalsIgnoreCase("age")) {
            return Integer.valueOf((int)(System.currentTimeMillis() / 1000) - session.getCreated());
        } else if (property.equalsIgnoreCase("idle")) {
            return Integer.valueOf((int)(System.currentTimeMillis() / 1000) - session.getLastActivity());
        } else if (property.equalsIgnoreCase("ttl")) {
            return Integer.valueOf(session.getExpiryTime() - (int)(System.currentTimeMillis() / 1000));
        } else if (property.equalsIgnoreCase("traffic")) {
            return Long.valueOf(session.getTraffic());
        } else if (property.equalsIgnoreCase("tracksStarted")) {
            return Integer.valueOf(session.getTracksStarted());
        } else if (property.equalsIgnoreCase("tracksCompleted")) {
            return Integer.valueOf(session.getTracksCompleted());
        } else {
            return session.getProperties().get(property);
        }
    }

    @SuppressWarnings("unchecked")
    public boolean test(String key, String op, final String testval, Object propval) {
        if (propval instanceof Service && (op.equals("=") || op.equals("!="))) {
            try {
                Class clazz = Class.forName(testval);
                return clazz.isAssignableFrom(propval.getClass()) == op.equals("=");
            } catch (ClassNotFoundException e) {
                return ((Service)propval).getContext().getServiceName().equals(testval) == op.equals("=");
            }
        } else if (propval instanceof InetAddress && (op.equals("=") || op.equals("!="))) {
            try {
                if (Character.isDigit(testval.charAt(0)) || testval.charAt(0)==':') {       // Test n.n.n.n or n.n.n.n/mask
                    int mask = 128;
                    int ix = testval.indexOf("/");
                    String t = testval;
                    if (ix >= 0) {
                        mask = Integer.parseInt(testval.substring(ix+1));
                        t = t.substring(0, ix);
                    }
                    InetAddress testaddr = InetAddress.getByName(t);
                    byte[] testbytes = testaddr.getAddress();
                    byte[] propbytes = ((InetAddress)propval).getAddress();
                    if (testbytes.length == propbytes.length) {
                        int maxmask = testbytes.length * 8;
                        mask = Math.min(mask, maxmask);
                        BigInteger testnum = new BigInteger(testbytes).shiftRight(maxmask - mask);
                        BigInteger propnum = new BigInteger(propbytes).shiftRight(maxmask - mask);
                        return op.equals("=")==testnum.equals(propnum);
                    } else {
                        return false;
                    }
                } else if (testval.startsWith("*.")) {
                    String propaddr = ((InetAddress)propval).getHostName();
                    return propaddr.endsWith(testval.substring(1));
                } else {
                    InetAddress[] testaddr = InetAddress.getAllByName(testval);
                    boolean equals = false;
                    for (int i=0;i<testaddr.length;i++) {
                        equals |= testaddr[i].equals(propval);
                    }
                    return equals==op.equals("=");
                }
            } catch (UnknownHostException e) {
                return false;
            }
        } else {
            if (propval instanceof InetAddress) {
                propval = ((InetAddress)propval).getHostName();
            }
            return super.test(key, op, testval, propval);
        }
    }

}
