package org.qtunes.core;

import java.net.*;
import java.util.*;

/**
 * A Session is a single login session and one must be in place for most {@link ControlMethod}
 * objects to be run. Each Session has a unique ID, a record of the Service that created it
 * and a number of "allowances", which determine what the Session can do. These can be checked
 * by calling the {@link #isAllowed} method.
 * </p><p>
 * When the session is created a set of Properties can be specified, which can be
 * retrieved later with the {@link #getProperties} method. Please see the API docs for that method
 * to get a list of standard properties.
 */
public abstract class Session {

    private volatile boolean expired;
    private volatile int tracksStarted, tracksCompleted;
    private volatile long traffic;

    /**
     * Return the unique numeric ID of this Session
     */
    public abstract int getId();

    /**
     * Return the Service that created this Session
     */
    public abstract Service getService();

    /**
     * Get the Properties specified when this Session was created. The list of recommended properties
     * is as follows - not all will be applicable for each type of Service
     * <table>
     * <tr><th>ip</th><td>The IP address or Hostname for the session, if it was created remotely</td></tr>
     * <tr><th>user</th><td>The username of the user that created the session</td></tr>
     * </table>
     */
    public abstract Map getProperties();

    /**
     * Mark the session as active as of this point in time
     */
    public abstract void touch();

    /**
     * Return the time this session was created, in seconds since epoch
     */
    public abstract int getCreated();

    /**
     * Return the last time any activity occurred on this session, in seconds since epoch
     */
    public abstract int getLastActivity();

    /**
     * Return the time this session should be expired, in seconds since epoch
     */
    public abstract int getExpiryTime();

    /**
     * Return whether this Session has the specified allowance. The allowances
     * depend on the {@link org.qtunes.auth.Authorizer} objects that have been registered with
     * the {@link org.qtunes.auth.Auth#addAuthorizer} method - an Authorizer that grants "*"
     * will allow any action, and passing the value "*" in as the test will
     * check for any allowance. The {@link #NONE} Session does not allow
     * anything.
     *
     * @param test the allowance we're testing for, or "*" for any allowance
     */
    public abstract boolean isAllowed(String test);

    /**
     * Return the Set of allowances for this Session
     */
    public abstract Collection<String> getAllowances();

    //------------------------------------------------------------------------

    /**
     * Return whether this session is expired
     */
    public boolean isExpired() {
        return expired;
    }

    /**
     * Mark this Session as expired. Should be called by {@link org.qtunes.auth.Auth#logout}.
     */
    public void expire() {
        expired = true;
    }

    /**
     * Increase the number of tracks started by this Session by 1
     */
    public void bumpTracksStarted() {
        tracksStarted++;
        touch();
    }

    /**
     * Increase the number of tracks completed by this Session by 1
     */
    public void bumpTracksCompleted() {
        tracksCompleted++;
        touch();
    }

    /**
     * Increase the amount of net traffic used by this session by the specified amount
     */
    public void bumpTraffic(long v) {
        traffic += v;
        touch();
    }

    /**
     * Return the amount of net traffic used by this Session
     */
    public long getTraffic() {
        return traffic;
    }

    /**
     * Return the number of tracks started by this Session
     */
    public int getTracksStarted() {
        return tracksStarted;
    }

    /**
     * Return the number of tracks completed by this Session
     */
    public int getTracksCompleted() {
        return tracksCompleted;
    }

    //------------------------------------------------------------------------------------

    /**
     * A Session placeholder to mark "no session". Returned from {@link #getCurrentSession} when 
     * no session is in place
     */
    public static final Session NONE = new Session() {
        public String toString() { return "<NOSESSION>"; }
        public int getId() { return 0; }
        public Map getProperties() { return Collections.EMPTY_MAP;}
        public Service getService() { return null; }
        public void touch() { }
        public int getCreated() { return 0; }
        public int getLastActivity() { return 0; }
        public int getExpiryTime() { return 0; }
        public boolean isAllowed(String key) { return "".equals(key); }
        public Collection<String> getAllowances() { return Collections.<String>emptySet(); }
    };

    private static final ThreadLocal<Session> currentSession = new InheritableThreadLocal<Session>() {
        protected Session initialValue() {
            return NONE;
        }
    };

    /**
     * Return the current Session for this Thread
     * @return the current Session or {@link #NONE} to indicate no active session
     */
    public synchronized static final Session getCurrentSession() {
        Session session = currentSession.get();
        if (session!=NONE && session.isExpired()) {
            currentSession.set(session = NONE);
        }
        return session;
    }

    /**
     * Set the current Session for this Thread
     * @param session the current Session, or {@link #NONE} to indicate no active session
     */
    public synchronized static final void setCurrentSession(Session session) {
        currentSession.set(session==null ? NONE : session);
    }

}
