package org.qtunes.daap;

import org.qtunes.core.*;
import org.qtunes.auth.*;
import org.qtunes.zeroconf.*;
import org.qtunes.player.*;
import java.util.*;
import java.util.concurrent.locks.*;
import java.io.*;
import java.net.*;
import java.security.*;

/**
 * Manage all the Remote applications that control Players
 * through the DACP protocol - also an Authorizer that is
 * used to verify new sessions for DACP and DAAP (where there is
 * no remote).
 */
class RemoteManager extends Thread implements Authorizer {

    private final ServiceContext context;
    private final List<String> prepending;
    private final List<PendingRemote> pending;
    private final List<Remote> paired;
    private final ReadWriteLock pairedlock;

    private boolean cancelled;
    private ServiceRegistrar serviceregistrar;

    RemoteManager(ServiceContext context) {
        super("qtunes-"+context.getServiceName()+"-RemoteManager");
        this.context = context;
        prepending = Collections.synchronizedList(new ArrayList<String>());
        pending = new ArrayList<PendingRemote>();
        pairedlock = new ReentrantReadWriteLock();
        paired = new ArrayList<Remote>();
        loadProperties();
        context.getService(Auth.class).addAuthorizer(this);
    }

    void setServiceRegistrar(ServiceRegistrar registrar) {
        this.serviceregistrar = registrar;
    }

    /**
     * Stop this service
     */
    synchronized void cancel() {
        context.getService(Auth.class).removeAuthorizer(this);
        this.cancelled = true;
        interrupt();
    }

    public Collection<String> getAllowances(Session session) {
        boolean match = false;
        if (session.getService() instanceof DaapServer) {
            DaapServer server = (DaapServer)session.getService();
            Map properties = session.getProperties();
            String guid = (String)properties.get("guid");
            if (guid == null) {
                match = server.isSharing();
            } else {
                pairedlock.readLock().lock();
                try {
                    for (Iterator<Remote> i = paired.iterator();!match && i.hasNext();) {
                        Remote remote = i.next();
                        if (remote.guid.equals(guid)) {
                            match = true;
                        }
                    }
                } finally {
                    pairedlock.readLock().unlock();
                }
            }
        }
        if (match) {
            return Arrays.asList(new String[] { "play", "playlists" });
        } else {
            return null;
        }
    }

    /**
     * Are there any remotes for the specified Player? Called from ServiceRegistrar
     */
    boolean hasRemotes(Player player) {
        String playername = player.getContext().getServiceName();
        pairedlock.readLock().lock();
        try {
            for (Iterator<Remote> i = paired.iterator();i.hasNext();) {
                Remote remote = i.next();
                if (remote.playername.equals(playername)) {
                    return true;
                }
            }
            return false;
        } finally {
            pairedlock.readLock().unlock();
        }
    }

    /**
     * Get the player for the specified guid. Called from DaapServer to
     * match a Player to a Session
     */
    Player getPlayer(String guid) {
        pairedlock.readLock().lock();
        try {
            for (Iterator<Remote> i = paired.iterator();i.hasNext();) {
                Remote remote = i.next();
                if (remote.guid.equals(guid)) {
                    return context.getService(Player.class, "name='"+remote.playername+"'");
                }
            }
            return null;
        } finally {
            pairedlock.readLock().unlock();
        }
    }

    public void run() {
        Set<String> oldpending = new HashSet<String>();
        Set<String> newpending;
        while (!cancelled) {
            ZeroConf zeroconf = context.getService(ZeroConf.class);
            newpending = new HashSet<String>();
            if (zeroconf != null) {
                ZCServiceInfo[] list = zeroconf.list("_touch-remote._tcp", 2000);
//                    System.out.println("touch-remote="+(list==null?"null":Arrays.asList(list).toString())+" old="+oldpending+" new="+newpending);
                for (int j=0;list!=null && j<list.length;j++) {
                    URL url = null;
                    ZCServiceInfo info = list[j];
                    Map properties = info.getProperties();
                    String name = (String)properties.get("DvNm");
                    String paircode = (String)properties.get("Pair");

                    boolean found = false;
                    PendingRemote pr = null;
                    synchronized(pending) {
                        for (Iterator<PendingRemote> i = pending.iterator();i.hasNext();) {
                            pr = i.next();
                            if (pr.name.equals(name) || pr.name.equals("*")) {
                                found = true;
                                InputStream in = null;
                                try {
                                    long uniqueid = DaapServer.getUniqueID(pr.player);
                                    String pairingcode = getPairingCode(paircode, pr.pin);

                                    url = new URL("http://"+info.getHost()+":"+info.getPort()+"/pair?pairingcode="+pairingcode+"&servicename="+hex(uniqueid));
                                    in = url.openConnection().getInputStream();
                                    ListBlock block = (ListBlock)Blocks.read(in, null);
                                    try {
                                        long cmpg = block.get(Blocks.cmpg).get();
                                        String cmnm = block.get(Blocks.cmnm).get();
                                        if (cmnm != null) {
                                            addPairedRemote(cmnm, pr.player, cmpg);
                                        }
                                    } catch (Exception e) {
                                        context.error("Failed to addPairedRemote: block="+block, e);
                                    }
                                } catch (Exception e) {
                                    if (e instanceof ConnectException && url != null) {
                                        context.error("Remote pairing for "+url, e);
                                    } else {
                                        context.error("Remote pairing", e);
                                    }
                                } finally {
                                    i.remove();
                                    if (in != null) {
                                        try { in.close(); } catch (Exception e) {}
                                    }
                                }
                            }
                        }
                        }
                    if (!found) {
                        newpending.add(name);
                    }
                }
            }
            for (Iterator<String> i = newpending.iterator();i.hasNext();) {
                String name = i.next();
                if (!oldpending.contains(name)) {
                    context.fireEvent("addPendingRemote", new Object[] { "name", name });
                    prepending.add(name);
                }
            }
            for (Iterator<String> i = oldpending.iterator();i.hasNext();) {
                String name = i.next();
                if (!newpending.contains(name)) {
                    context.fireEvent("delPendingRemote", new Object[] { "name", name });
                    prepending.remove(name);
                }
            }
            oldpending = newpending;
            try {
                Thread.sleep(1000);
            } catch (InterruptedException e) {}
        }
    }

    List<String> getPendingRemotes() {
        return Collections.unmodifiableList(prepending);
    }

    /**
     * Pair a new remote
     */
    void addRemote(String name, Player player, String pin) {
        removeRemote(name, player);
        pin = ("000"+pin).substring(pin.length()-1);
        pending.add(new PendingRemote(name, player, pin));
        synchronized(this) {
            interrupt();
        }
    }

    /**
     * Remove a previously paired remote
     */
    void removeRemote(String name, Player player) {
        String playername = player.getContext().getServiceName();
        boolean changed = false;
        pairedlock.writeLock().lock();
        try {
            for (Iterator<Remote> i = paired.iterator();!changed && i.hasNext();) {
                Remote remote = i.next();
                if (remote.playername.equals(playername) && remote.name.equals(name)) {
                    i.remove();
                    context.fireEvent("unregisterRemote", new Object[] { "name", name, "player", playername, "guid", remote.guid });
                    serviceregistrar.update();
                    saveProperties();
                }
            }
        } finally {
            pairedlock.writeLock().unlock();
            
        }
    }

    /**
     * Add a completely paired remote 
     */
    private void addPairedRemote(String name, Player player, long guid) {
        String guidstring = "0x"+hex(guid);
        String playername = player.getContext().getServiceName();
        pairedlock.writeLock().lock();
        try {
            paired.add(new Remote(playername, guidstring, name));
        } finally {
            pairedlock.writeLock().unlock();
        }
            
        context.fireEvent("registerRemote", new Object[] { "name", name, "player", playername, "guid", guidstring });
        serviceregistrar.update();
        saveProperties();
    }

    //-------------------------------------------------------------------------

    private String hex(long l) {
        String v = Long.toHexString(l).toUpperCase();
        return ("000000000000000"+v).substring(v.length() - 1);
    }

    private void loadProperties() {
        String[] propertynames = context.getPropertyNames();
        for (int i=0;i<propertynames.length;i++) {
            if (propertynames[i].startsWith("remote.")) {
                Remote remote = Remote.read(context.getProperty(propertynames[i]));
                if (remote != null) {
                    paired.add(remote);
                }
            }
        }
    }

    private void saveProperties() {
        String[] propertynames = context.getPropertyNames();
        for (int i=0;i<propertynames.length;i++) {
            if (propertynames[i].startsWith("remote.")) {
                context.putProperty(propertynames[i], null);
            }
        }
        int c = 0;
        for (Iterator<Remote> i = paired.iterator();i.hasNext();c++) {
            Remote remote = i.next();
            context.putProperty("remote."+c, remote.write());
        }
    }

    private String getPairingCode(String paircode, String pin) {
        try {
            MessageDigest digest = MessageDigest.getInstance("MD5");
            digest.update(paircode.getBytes("ISO-8859-1"));
            for (int i=0;i<4;i++) {
                digest.update((byte)pin.charAt(i));
                digest.update((byte)0);
            }
            byte[] data = digest.digest();
            StringBuffer sb = new StringBuffer(data.length * 2);
            for (int i=0;i<data.length;i++) {
                int v = data[i] & 0xFF;
                sb.append((v < 16 ? "0" : "") + Integer.toHexString(v).toUpperCase());
            }
            return sb.toString();
        } catch (IOException e) {
            throw new RuntimeException(e);
        } catch (NoSuchAlgorithmException e) {
            throw new RuntimeException(e);
        }
    }

    private static class PendingRemote {
        final String name, pin;
        final Player player;
        PendingRemote(String name, Player player, String pin) {
            this.name = name;
            this.player = player;
            this.pin = pin;
        }
    }

    private static class Remote {
        final String name, guid, playername;
        Remote(String playername, String guid, String name) {
            this.guid = guid;
            this.name = name;
            this.playername = playername;
        }
        static Remote read(String value) {
            int ix1 = value.indexOf(':');
            int ix2 = value.indexOf(':', ix1+1);
            if (ix2 >= 0) {
                String player = value.substring(0, ix1);
                String guid = value.substring(ix1+1, ix2);
                String name = value.substring(ix2+1);
                return new Remote(player, guid, name);
            } else {
                return null;
            }
        }
        String write() {
            return toString();
        }
        public String toString() {
            return playername+":"+guid+":"+name;
        }
    }

}
