package org.qtunes.db.spi.simple;

import org.qtunes.core.*;
import org.qtunes.db.*;
import java.nio.*;
import java.nio.channels.*;
import java.util.concurrent.locks.*;
import java.io.*;
import java.util.*;
import java.lang.reflect.Type;

/**
 * A type of Store that uses a simple but very fast custom file format
 * on disk to store Track data. The file format is read-only, so a
 * merge will create a brand new file to replace the existing one. For
 * infrequent merges this is a very efficient option.
 */
class FileStore implements Store {

    private final Record[] threadrecords;
    private final ThreadLocal<Record> threadlocalrecord = new ThreadLocal<Record>();
    private static final int BUFSIZE = 65536;
    private File file;
    private volatile FileChannel channel;
    private int maxtrack, revision;
    private int[] offsets;
    private BitSet tracks;
    private SimpleIndices indices;
    private final ReadWriteLock dblock;

    private final int numfields = Field.ALLFIELDS.length;

    FileStore() {
	dblock = new ReentrantReadWriteLock();
        threadrecords = new Record[Runtime.getRuntime().availableProcessors()];
	for (int i=0;i<threadrecords.length;i++) {
	    threadrecords[i] = new Record(i);
	}
    }

    public boolean isValid() {
        return file != null;
    }

    public void setFile(String filename) {
        if (filename == null) {
            throw new IllegalArgumentException("No file specified");
        }
        this.file = new File(filename);
        this.indices = new SimpleIndices(this);
    }

    public BitSet getTracks() {
        load();
        return tracks;
    }

    public int[] getTracks(String sort, final String filter) {
        if (indices.isDirty()) {
            indices.rebuild();
        }
        BitSet t;
        if (filter == null || filter.length()==0) {
            t = getTracks();
        } else {
            t = indices.getFilteredTracks(filter);
        }
        List<Integer> list = new ArrayList<Integer>(t.size());
        for (int i=t.nextSetBit(0);i>=0;i=t.nextSetBit(i+1)) {
            list.add(Integer.valueOf(i));
        }
        if (sort != null) {
            Comparator<Integer> comparator = indices.getComparator(sort);
            Collections.sort(list, comparator);
        }
        int[] out = new int[list.size()];
        for (int i=0;i<out.length;i++) {
            out[i] = list.get(i).intValue();
        }
        return out;
    }

    public int getRevision() {
        load();
        return revision;
    }

    /**
     * The file format is as follows:
     * 
     * maxtrack         int                    highest track index in the DB
     * revision         int                    a revision number - should increase for each save
     * offsets          int[maxtrack]          offsets from start of file to each track record
     * tracks           trackrecord[maxtrack]  track record for each track
     *
     * "trackrecord" is as follows:
     *
     * fieldoffsets     short[numfields]        offsets into track record for each field
     * fieldvalues      fieldrecord[numfields]  field record for each field
     * 
     * "fieldrecord" is very simple - for int, long, float or boolean it's the value as
     * written by {@link ByteBuffer}. For Strings it's the number of bytes in the UTF-8 String,
     * (stored as a short) followed by the UTF-8 bytes. For int[] or byte[] it's the length
     * of the array (as a short) followed by the array values as written by {@link ByteBuffer}
     *
     * This is a fairly simple format but it means a field lookup for a track involves two
     * dereferences, for track then field. All field data is grouped nearby so likely will
     * already be off disk and in the cache. It's extremely quick and quite space efficient,
     * as the only wasted space is in the fieldoffsets array for absent fields.
     *
     * Updates are done by rebuilding this file completely, which is why it's only done on a commit
     */
    private void load() {
        if (channel == null) {
            if (file == null) {
                throw new IllegalStateException("FileStore filename not specified");
            }
	    try {
		dblock.readLock().lock();
		if (file.canRead()) {
		    try {
			file = file.getCanonicalFile();
			FileInputStream tin = new FileInputStream(file);
			channel = tin.getChannel();
			ByteBuffer buf = ByteBuffer.allocate(8);
			channel.read(buf);
			buf.flip();

			maxtrack = buf.getInt();
			offsets = new int[maxtrack+1];
			offsets[maxtrack] = (int)channel.size();
			revision = buf.getInt();
			tracks = new BitSet(maxtrack);
			buf = ByteBuffer.allocate(4 * maxtrack);
			channel.read(buf);
			buf.flip();
			for (int i=0;i<maxtrack;i++) {
			    offsets[i] = buf.getInt();
			    if (offsets[i] != 0) {
				tracks.set(i);
			    }
			}
		    } catch (IOException e) {
			throw new IllegalStateException(e);
		    }
		} else {
		    tracks = new BitSet();
		    maxtrack = 0;
		}
	    } finally {
	        dblock.readLock().unlock();
	    }
        }
    }

    /**
     * Load the ByteBuffer containing the field info for the specified track.
     * Note the MappedByteBuffer is NOT used - this class has its uses but
     * for long term use over a single file like this I'm worried the complications
     * it brings to garbage collection are too much.
     */
    private ByteBuffer getRecord(int track) {
	ByteBuffer out = null;
        try {
            load();
	    if (track >= 0 && track < maxtrack) {
		int j = track;
		while (j < maxtrack && offsets[++j]==0);	// Skip deleted records
                final int recordoff = offsets[track];
		if (recordoff > 0) {
		    final int recordlen = offsets[j] - recordoff;

		    Record record = leaseRecord();			// Thread local
		    ByteBuffer recbuf = record.buf;
		    if (recbuf == null || recordoff < record.offset || recordoff + recordlen > record.offset + record.buf.limit()) {
			if (recbuf == null) {
			    recbuf = record.buf = ByteBuffer.allocate(BUFSIZE);
			}
			recbuf.clear();
			try {
			    dblock.readLock().lock();
			    recbuf.limit(channel.read(recbuf, recordoff));
			} finally {
			    dblock.readLock().unlock();
			}
			record.offset = recordoff;
		    }
		    int limit = recbuf.limit();
		    recbuf.position(recordoff - record.offset);
		    recbuf.limit(recbuf.position() + recordlen);
		    out = (ByteBuffer)recbuf.slice();
		    recbuf.limit(limit);
		}
	    }
        } catch (IOException e) {
            throw new IllegalStateException(e);
        }
	return out;
    }

    public Object get(int track, Field field) {
        ByteBuffer record = getRecord(track);
	if (record != null) {
	    try {
		int fieldoff = record.getShort(field.getIndex() * 2);
		if (fieldoff > 0) {
		    Class type = field.getType();
		    record.position(fieldoff);
		    if (type == Integer.class) {
			return Integer.valueOf(record.getInt());
		    } else if (type == Long.class) {
			return Long.valueOf(record.getLong());
		    } else if (type == Float.class) {
			return Float.valueOf(record.getFloat());
		    } else if (type == Boolean.class) {
			return Boolean.TRUE;
		    } else if (type == String.class) {
			int len = record.getShort();
			byte[] array = new byte[len];
			try {
			    record.get(array);
			    return new String(array, "UTF-8");
			} catch (Exception e) {
			    throw new IllegalStateException(e);
			}
		    } else if (type == byte[].class) {
			int len = record.getShort();
			byte[] array = new byte[len];
			record.get(array);
			return array;
		    } else if (type == int[].class) {
			int len = record.getShort();
			int[] array = new int[len];
			for (int i=0;i<array.length;i++) {
			    array[i] = record.getInt();
			}
			return array;
		    } else {
			throw new IllegalStateException(field+" "+type);
		    }
		}
	    } catch (RuntimeException e) {
	        System.out.println("track="+track+"/"+maxtrack+" off="+offsets[track]+" off+1="+offsets[track+1]+" len="+(offsets[track+1]-offsets[track])+" buflen="+record.position()+"/"+record.limit()+"/"+record.capacity());
		throw e;
	    } finally {
	        freeRecord();
	    }
	}
	return null;
    }

    public boolean isDeleted(int track) {
        return !tracks.get(track);
    }

    public void delete(int track) {
	BitSet tracks = this.tracks;
        if (tracks != null && track < maxtrack) {
            tracks.clear(track);
            offsets[track] = 0;
        }
    }

    public BitSet getFields(int track) {
        ByteBuffer record = getRecord(track);
	BitSet fields = null;
	if (record != null) {
	    try {
		fields = new BitSet(numfields);
		for (int i=0;i<numfields;i++) {
		    if (record.getShort(i*2) != 0) {
			fields.set(i);
		    }
		}
	    } finally {
	        freeRecord();
	    }
	}
	return fields;
    }

    /**
     * Mark the record used by this thread as free
     */
    private void freeRecord() {
        synchronized(threadrecords) {
	    Record rec = threadlocalrecord.get();
	    rec.thread = null;
	    threadlocalrecord.remove();
	    threadrecords.notify();
	}
    }

    /**
     * Return a record which is exclusively for use by this thread. Keep array
     * of n Records which we can reuse per each thread, and use threadlocalrecord
     * to tell each thread which one it has. Better than new on each threadlocal
     * as this system creates many threads. If no free record is available we
     * will wait.
     */
    private Record leaseRecord() {
	synchronized(threadrecords) {
	    while (true) {
		for (int i=0;i<threadrecords.length;i++) {
		    Record r = threadrecords[i];
		    if (r.thread == null) {
			r.thread = Thread.currentThread();
			threadlocalrecord.set(r);
			return r;
		    }
		}
		try {
		    threadrecords.wait();
		} catch (InterruptedException e) { }
	    }
	}
    }

    /**
     * Merge the tracks from the specified Store into this one.
     * This makes a new database file and copies the merged
     * records into that, then replaces the current file. It's
     * a slowish operation but infrequently called.
     */
    public void merge(Store store) throws IOException {
        File temp = File.createTempFile(file.getName(), null, file.getAbsoluteFile().getParentFile());
        FileOutputStream sout = new FileOutputStream(temp);
        FileChannel out = sout.getChannel();
        BitSet newtracks = store.getTracks();
	int outposition = 0;
        int maxtrack = Math.max(getTracks().length(), newtracks.length());
        ByteBuffer header = ByteBuffer.allocate((maxtrack + 2) * 4);
        header.putInt(maxtrack);
        header.putInt(++revision);
        header.rewind();
        out.write(header);
	outposition += header.limit();

        ByteBuffer work = ByteBuffer.allocate(BUFSIZE);
        int delcount = 0, newcount = 0, oldcount = 0;
	for (int i=0;i<maxtrack;i++) {
	    if (store.isDeleted(i)) {
		delcount++;
	    } else if (newtracks.get(i)) {
		newcount++;
		createRecord(store, i, work);
		header.putInt(8 + (i*4), outposition);
		out.write(work);
		outposition += work.limit();
	    } else if (offsets != null && offsets[i] != 0) {
		oldcount++;
		int j = i;
		while (i < maxtrack && offsets[++j]==0);
		header.putInt(8 + (i*4), outposition);
		int tlen = offsets[j] - offsets[i];
		int toff = 0;
		work.clear();
		while (toff < tlen) {
		    if (false) {
			// Doesn't work!!! out.position() is not incremented, resulting
			// file is correct size when closed, but corrupt.
			// Failing under Linux 1.6.0_24.
			toff += channel.transferTo(offsets[i] + toff, tlen - toff, out);
		    } else {
			work.limit(tlen - toff);
			toff += channel.read(work, offsets[i] + toff);
			work.flip();
			out.write(work);
		    }
		}
		outposition += tlen;
	    }
	}

	// Move DB file - now we writelock and invalidate previous records. Probably
	// a microscopic window for threadrecords to wind up with wrong offset here.
	dblock.writeLock().lock();
	try {
	    for (int i=0;i<threadrecords.length;i++) {
		threadrecords[i].offset = Integer.MAX_VALUE;
	    }
	    header.rewind();
	    out.position(0);
	    out.write(header);
	    out.close();
	    if (channel != null) {
		channel.close();
		channel = null;
	    }
	    tracks = null;
	    if (!temp.renameTo(file)) {
		throw new IOException("Unable to rename "+temp+" to "+file);
	    }
	    indices.setDirty();
        } finally {
	    dblock.writeLock().unlock();
	}
    }

    private void createRecord(Store store, int track, ByteBuffer buf) {
        buf.clear();
        BitSet fields = store.getFields(track);
        for (int i=0;i<numfields;i++) {
            buf.putShort((short)0);
        }

        for (int i=0;i<numfields;i++) {
            if (fields.get(i)) {
                buf.putShort(i * 2, (short)buf.position());
                Field field = Field.ALLFIELDS[i];
                Object val = store.get(track, field);
                Type type = field.getType();
                if (type == Integer.class) {
                    buf.putInt(((Integer)val).intValue());
                } else if (type == Long.class) {
                    buf.putLong(((Long)val).longValue());
                } else if (type == Float.class) {
                    buf.putFloat(((Float)val).floatValue());
                } else if (type == Boolean.class) {
                } else if (type == byte[].class) {
                    byte[] array = (byte[])val;
                    buf.putShort((short)array.length);
                    buf.put(array);
                } else if (type == int[].class) {
                    int[] array = (int[])val;
                    buf.putShort((short)array.length);
                    for (int k=0;k<array.length;k++) {
                        buf.putInt(array[k]);
                    }
                } else if (type == String.class) {
                    try {
                        byte[] array = ((String)val).getBytes("UTF-8");
                        buf.putShort((short)array.length);
                        buf.put(array);
                    } catch (IOException e) {
                        throw new RuntimeException(e);
                    }
                } else {
                    throw new IllegalStateException();
                }
            }
        }
        buf.flip();
    }

    private static class Record {
	final int index;
        ByteBuffer buf;
	int offset;
	Thread thread;
	Record(int i) {
	    this.index = i;
	}
	public String toString() {
	    return "[rec"+index+"]";
	}
    }

}
