package org.qtunes.db.spi.simple;

import org.qtunes.db.*;
import java.io.*;
import java.util.*;
import java.text.*;
import java.util.concurrent.locks.*;
import org.apache.lucene.analysis.*;
import org.apache.lucene.document.*;
import org.apache.lucene.document.Field;
import org.apache.lucene.index.*;
import org.apache.lucene.search.*;
import org.apache.lucene.collation.*;
import org.apache.lucene.store.*;
import org.apache.lucene.queryParser.*;
import org.apache.lucene.util.*;

class LuceneStore implements Store {

    private static final Version VERSION = Version.LUCENE_32;
    private Directory directory;
    private final Analyzer analyzer;
    private int revision;
    private IndexReader index;
    private IndexSearcher searcher;
    private IndexConverter indexconverter;
    private ReentrantReadWriteLock dblock = new ReentrantReadWriteLock();       // dblock, for tracks

    public LuceneStore() {
        analyzer = new ReusableAnalyzerBase() {
            protected TokenStreamComponents createComponents(final String fieldname, final Reader reader) {
                Tokenizer src = new LowerCaseTokenizer(VERSION, reader);
                TokenStream sink = new ASCIIFoldingFilter(src);
                return new TokenStreamComponents(src, sink);
            }
        };
    }

    static final int NORMALIZATIONMASK = (1<<Character.DECIMAL_DIGIT_NUMBER)|(1<<Character.LETTER_NUMBER)|(1<<Character.LOWERCASE_LETTER)|(1<<Character.MATH_SYMBOL)|(1<<Character.OTHER_LETTER)|(1<<Character.OTHER_NUMBER)|(1<<Character.OTHER_SYMBOL)|(1<<Character.TITLECASE_LETTER)|(1<<Character.UPPERCASE_LETTER)|(1<<Character.SPACE_SEPARATOR);

    /**
     * Given a String, return the String that it should be sorted on. Strips leading "The",
     * diacritics, punctuation and spaces.
     */
    private static String createSortString(String o) {
        o = java.text.Normalizer.normalize((String)o.toLowerCase().trim(), java.text.Normalizer.Form.NFD);
        if (o.startsWith("the ")) {
            o = o.substring(4);
        }
        StringBuffer sb = new StringBuffer(o.length());
        for (int i=0;i<o.length();i++) {
            char c = o.charAt(i);
            int type = Character.getType(c);
            if (((1<<type)&NORMALIZATIONMASK) != 0) {
                sb.append(c);
            }
        }
        return sb.toString();
    }

    public synchronized void setFile(String file) {
        dblock.readLock().lock();
        try {
            directory = FSDirectory.open(new File(file));
            if (!IndexReader.indexExists(directory)) {
                IndexWriter writer = getIndexWriter();
                writer.close();
            }
        } catch (Exception e) {
            if (e instanceof RuntimeException) {
                throw (RuntimeException)e;
            } else {
                throw new RuntimeException(e);
            }
        } finally {
            dblock.readLock().unlock();
        }
        load();
    }

    public boolean isValid() {
        return directory != null;
    }

    private synchronized void load() {
        dblock.writeLock().lock();
        try {
            index = IndexReader.open(directory);
            searcher = new IndexSearcher(directory);
            int max = index.maxDoc();
            indexconverter = new IndexConverter(max);
            for (int i=0;i<max;i++) {
                Document doc = index.document(i);
                int ext = Integer.valueOf(((Field)doc.getFieldable("index")).stringValue());
                indexconverter.setMapping(i, ext);
            }
        } catch (Exception e) {
            if (e instanceof RuntimeException) {
                throw (RuntimeException)e;
            } else {
                throw new RuntimeException(e);
            }
        } finally {
            dblock.writeLock().unlock();
        }
        revision++;
    }

    public BitSet getTracks() {
        dblock.readLock().lock();
        try {
            BitSet bs = new BitSet();
            for (int i=0;i<indexconverter.getMaxInternalIndex();i++) {
                int extix = indexconverter.getExternalIndex(i);
                if (extix >= 0) {
                    bs.set(extix);
                }
            }
            return bs;
        } catch (Exception e) {
            if (e instanceof RuntimeException) {
                throw (RuntimeException)e;
            } else {
                throw new RuntimeException(e);
            }
        } finally {
            dblock.readLock().unlock();
        }
    }

    public void delete(int track) {
        dblock.readLock().lock();
        try {
            indexconverter.removeExternalIndex(track);
        } finally {
            dblock.readLock().unlock();
        }
    }

    public boolean isDeleted(int track) {
        dblock.readLock().lock();
        try {
            return indexconverter.getInternalIndex(track) < 0;
        } finally {
            dblock.readLock().unlock();
        }
    }

    public Object get(int track, org.qtunes.db.Field field) {
        dblock.readLock().lock();
        try {
            int ix = indexconverter.getInternalIndex(track);
            if (ix >= 0) {
                Document doc = index.document(ix);
                Fieldable f = doc.getFieldable(field.getName());
                if (f == null) {
                    if (field == org.qtunes.db.Field.Title || field == org.qtunes.db.Field.File) {
                        throw new Error("track="+track+" ix="+ix);
                    }
                    return null;
                } else if (field.getType() == String.class) {
                    return ((Field)f).stringValue();
                } else if (field.getType() == Integer.class) {
                    if (f instanceof NumericField) {
                        return (Integer)((NumericField)f).getNumericValue();
                    } else {
                        return Integer.valueOf(((NumericField)f).stringValue());
                    }
                } else if (field.getType() == Long.class) {
                    if (f instanceof NumericField) {
                        return (Long)((NumericField)f).getNumericValue();
                    } else {
                        return Long.valueOf(((Field)f).stringValue());
                    }
                } else if (field.getType() == Float.class) {
                    if (f instanceof NumericField) {
                        return (Float)((NumericField)f).getNumericValue();
                    } else {
                        return Float.valueOf(((Field)f).stringValue());
                    }
                } else if (field.getType() == Boolean.class) {
                    return Boolean.TRUE;
                } else {
                    throw new ClassCastException(field.getType().getName());
                }
            } else {
                return null;
            }
        } catch (Exception e) {
            if (e instanceof RuntimeException) {
                throw (RuntimeException)e;
            } else {
                throw new RuntimeException(e);
            }
        } finally {
            dblock.readLock().unlock();
        }
    }

    public BitSet getFields(int track) {
        dblock.readLock().lock();
        try {
            int ix = indexconverter.getInternalIndex(track);
            BitSet set = new BitSet();
            if (ix >= 0) {
                Document doc = index.document(ix);
                List<Fieldable> l = doc.getFields();
                for (int i=0;i<l.size();i++) {
                    org.qtunes.db.AbstractField f = org.qtunes.db.Field.getByName(l.get(i).name());
                    if (f instanceof org.qtunes.db.Field) {
                        set.set(((org.qtunes.db.Field)f).getIndex());
                    }
                }
            }
            return set;
        } catch (Exception e) {
            if (e instanceof RuntimeException) {
                throw (RuntimeException)e;
            } else {
                throw new RuntimeException(e);
            }
        } finally {
            dblock.readLock().unlock();
        }
    }

    public int getRevision() {
        return revision;
    }

    public void merge(Store store) throws Exception {
        dblock.readLock().lock();
        try {
            IndexReader newindex = IndexReader.open(directory, false);
            for (int i=0;i<indexconverter.getMaxInternalIndex();i++) {
                int ex = indexconverter.getExternalIndex(i);
                if (ex >= 0 && indexconverter.getInternalIndex(ex) < 0) {
                    newindex.deleteDocument(i);
                }
            }
            BitSet bs = store.getTracks();
            for (int i = bs.nextSetBit(0); i >= 0; i = bs.nextSetBit(i+1)) {
                int ix = indexconverter.getInternalIndex(i);
                if (ix >= 0) {
                    newindex.deleteDocument(ix);
                }
            }
            boolean optimize = newindex.hasDeletions();
            newindex.close();

            IndexWriter writer = getIndexWriter();
            for (int i = bs.nextSetBit(0); i >= 0; i = bs.nextSetBit(i+1)) {
                Document doc = new Document();
                doc.add(new Field("index", Integer.toString(i), Field.Store.YES, Field.Index.NOT_ANALYZED));
                BitSet fields = store.getFields(i);
                for (int j = fields.nextSetBit(0); j >= 0; j = fields.nextSetBit(j+1)) {
                    org.qtunes.db.Field qtfield = org.qtunes.db.Field.ALLFIELDS[j];
                    Object o = store.get(i, qtfield);
                    Fieldable field = null;
                    if (o == Boolean.TRUE) {
                        o = Integer.valueOf(1);
                    } else if (o == Boolean.FALSE) {
                        o = null;
                    }
                    if (o != null) {
                        Field.Index fieldindex = qtfield.getSearch() == org.qtunes.db.Field.Search.NONE ? Field.Index.NO : qtfield.getSearch() == org.qtunes.db.Field.Search.INDEX ? Field.Index.NOT_ANALYZED : Field.Index.ANALYZED;
                        Field.Store fieldstore = Field.Store.YES;

                        if (o instanceof String) {
                            field = new Field(qtfield.getName(), (String)o, fieldstore, fieldindex);
                        } else {
                            NumericField f = new NumericField(qtfield.getName(), fieldstore, fieldindex != Field.Index.NO);
                            if (o instanceof Integer) {
                                f.setIntValue(((Number)o).intValue());
                            } else if (o instanceof Long) {
                                f.setLongValue(((Number)o).longValue());
                            } else {
                                f.setFloatValue(((Number)o).floatValue());
                            }
                            field = f;
                        }
                        doc.add(field);
                        if (field.isTokenized() && o instanceof String) {
                            // Can't sort on Tokenized fields
                            String sortvalue = createSortString((String)o);
                            field = new Field(qtfield.getName()+"-Sort", sortvalue, Field.Store.NO, Field.Index.NOT_ANALYZED);
                            doc.add(field);
                        }
                    }
                }
                writer.addDocument(doc);
            }
            if (optimize) {
                writer.optimize();
            }
            writer.close();
        } finally {
            dblock.readLock().unlock();
        }
        load();
    }

    public int[] getTracks(String sort, String filter) {
        dblock.readLock().lock();
        if (index.maxDoc() == 0) {
            return new int[0];
        }
        try {
            Sort sorter = null;
            if (sort == null || sort.length() == 0) {
                sort = "+album";
            }
            boolean reverse = sort.charAt(0) == '-';
            sort = sort.substring(1);
            if (sort.equals("album")) {
                sorter = new Sort(new SortField("Album-Sort", SortField.STRING, reverse), new SortField("Disc", SortField.INT), new SortField("Track", SortField.INT), new SortField("Artist-Sort", SortField.STRING), new SortField("Title-Sort", SortField.STRING));
            } else if (sort.equals("title")) {
                sorter = new Sort(new SortField("Title-Sort", SortField.STRING, reverse), new SortField("Artist", SortField.STRING), new SortField("Album", SortField.STRING), new SortField("Disc", SortField.INT), new SortField("Track", SortField.INT));
            } else if (sort.equals("artist")) {
                sorter = new Sort(new SortField("Artist-Sort", SortField.STRING, reverse), new SortField("Album-Sort", SortField.STRING, reverse), new SortField("Disc", SortField.INT), new SortField("Track-Sort", SortField.INT), new SortField("Title-Sort", SortField.STRING));
            } else if (sort.equals("albumartist")) {
                sorter = new Sort(new SortField("AlbumArtist-Sort", SortField.STRING, reverse), new SortField("Album-Sort", SortField.STRING, reverse), new SortField("Disc", SortField.INT), new SortField("Track", SortField.INT), new SortField("Title-Sort", SortField.STRING));
            } else if (sort.equals("group")) {
                sorter = new Sort(new SortField("Group-Sort", SortField.STRING, reverse), new SortField("Album-Sort", SortField.STRING), new SortField("Disc", SortField.INT), new SortField("Track", SortField.INT), new SortField("Artist-Sort", SortField.STRING), new SortField("Title-Sort", SortField.STRING));
            } else {
                throw new IllegalArgumentException("Unknown sort \""+sort+"\"");
            }

            Query query;
            if (filter == null || filter.length() == 0) {
                query = new MatchAllDocsQuery();
            } else {
                List<String> list = new ArrayList<String>();
                for (int i=0;i<org.qtunes.db.Field.ALLFIELDS.length;i++) {
                    org.qtunes.db.Field qtfield = org.qtunes.db.Field.ALLFIELDS[i];
                    if (qtfield.getSearch() == org.qtunes.db.Field.Search.SEARCH) {
                        list.add(qtfield.getName());
                    }
                }
                MultiFieldQueryParser mfqp = new MultiFieldQueryParser(VERSION, list.toArray(new String[0]), analyzer);
                mfqp.setDefaultOperator(QueryParser.Operator.AND);
                query = mfqp.parse(filter);
            }

            searcher = new IndexSearcher(directory);
            TopDocs rs = searcher.search(query, searcher.getIndexReader().numDocs(), sorter);
            int[] out = new int[rs.scoreDocs.length];
            for (int i=0;i<out.length;i++) {
                out[i] = rs.scoreDocs[i].doc;
            }
            for (int i=0;i<out.length;i++) {
                out[i] = indexconverter.getExternalIndex(rs.scoreDocs[i].doc);
            }
            return out;
        } catch (Exception e) {
            if (e instanceof RuntimeException) {
                throw (RuntimeException)e;
            } else {
                throw new RuntimeException(e);
            }
        } finally {
            dblock.readLock().unlock();
        }
    }

    private IndexWriter getIndexWriter() throws Exception {
        IndexWriterConfig config = new IndexWriterConfig(VERSION, analyzer);
        IndexWriter writer = new IndexWriter(directory, config);
        return writer;
    }

    private class IndexConverter {
        private int[] i2e, e2i;
        private int maxext;

        IndexConverter(int max) {
            i2e = new int[max];
            e2i = new int[max];
            for (int i=0;i<max;i++) {
                e2i[i] = i2e[i] = -1;
            }
        }

        void setMapping(int internal, int external) {
            i2e[internal] = external;
            if (external >= e2i.length) {
                int[] z = new int[external + (external>>2)];
                System.arraycopy(e2i, 0, z, 0, e2i.length);
                for (int j=e2i.length;j<z.length;j++) {
                    z[j] = -1;
                }
                e2i = z;
            }
            if (external > maxext) {
                maxext = external;
            }
            e2i[external] = internal;
        }

        void close() {
            if (e2i.length > maxext) {
                int[] z = new int[maxext];
                System.arraycopy(e2i, 0, z, 0, maxext);
                e2i = z;
            }
        }

        int getExternalIndex(int ix) {
            return ix >= i2e.length ? -1 : i2e[ix];
        }

        int getInternalIndex(int ix) {
            return ix >= e2i.length ? -1 : e2i[ix];
        }

        int getMaxInternalIndex() {
            return i2e.length;
        }

        int getMaxExternalIndex() {
            return e2i.length;
        }

        void removeExternalIndex(int ix) {
            e2i[ix] = -1;
        }
    }

}
