package org.qtunes.db.spi.simple;

import org.qtunes.core.*;
import org.qtunes.db.*;
import java.util.*;
import java.io.*;

/**
 * A Playlist which determines it's tracks based on a filter. The
 * filter uses a {@link TrackMatcher}, so you can for example do
 * <code>Album=~"Buzzin' Fly.*"</code> or <code>(Artist="Postal Service" or
 * artist=~"Death Cab.*") and Year > "2000"</code>.
 */
public class SmartPlaylistImpl implements SmartPlaylist {

    private final DatabaseImpl db;
    private TrackMatcher matcher;
    private transient List<Track> tracklist;
    private transient Map<String,Playlist> cache = new HashMap<String,Playlist>();
    private transient int lastdbrevision, revision, uniqueid;
    private transient String name;

    SmartPlaylistImpl(DatabaseImpl db, int uniqueid, String filter) {
        this.db = db;
        this.uniqueid = uniqueid;
        this.matcher = new TrackMatcher(filter);
    }

    public void setQuery(String query) {
        this.matcher = new TrackMatcher(query);
        cache.clear();
        tracklist = null;
        revision++;
        if (name != null) {
            db.putPlaylist(name, this);         // Save on any change
        }
    }

    public void setName(String name) {
        this.name = name;
    }

    public String getName() {
        return name;
    }

    public int getUniqueID() {
        return uniqueid;
    }

    public int getRevision() {
        return (lastdbrevision << 8) + revision; 
    }

    public int hashCode() {
        return matcher.hashCode();
    }

    public boolean hasTrack(Track track) {
        return matcher.matches(track);
    }

    public boolean equals(Object o) {
        return o instanceof SmartPlaylistImpl ? ((SmartPlaylistImpl)o).matcher.equals(matcher) : false;
    }

    public String toString() {
        return matcher.toString();
    }

    public String getQuery() {
        return matcher.getFilter();
    }

    public synchronized List<Track> getTracks() {
        if (db.getRevision() > lastdbrevision) {
            cache.clear();
            tracklist = null;
        }
        if (tracklist == null) {
            lastdbrevision = db.getRevision();
            Collection<Track> alltracks = db.getTracks();
            List<Track> worklist = new ArrayList<Track>();
            for (Iterator<Track> i = alltracks.iterator();i.hasNext();) {
                Track track = i.next();
                if (hasTrack(track)) {
                    worklist.add(track);
                }
            }
            tracklist = Collections.unmodifiableList(worklist);
        }
        return tracklist;
    }

    public synchronized Playlist sort(final String sort, final String filter) {
        String key = sort+"/"+filter;
        Playlist pl = cache.get(key);
        if (pl == null) {
            cache.put(key, pl = new Playlist() {
                List<Track> filteredlist;
                int revision;
                public synchronized List<Track> getTracks() {
                    if (filteredlist == null) {
                        revision = db.getRevision();
                        List<Track> alltracks = db.getTracks(sort, filter);
                        if (matcher.isUniversal()) {
                            filteredlist = Collections.unmodifiableList(alltracks);
                        } else {
                            List<Track> worklist = new ArrayList<Track>();
                            for (int i=0;i<alltracks.size();i++) {
                                Track track = alltracks.get(i);
                                if (hasTrack(track)) {
                                    worklist.add(track);
                                }
                            }
                            filteredlist = Collections.unmodifiableList(worklist);
                        }
                    }
                    return filteredlist;
                }
                public int size() {
                    return getTracks().size();
                }
                public String getName() {
                    return SmartPlaylistImpl.this.getName();
                } 
                public void setName(String name) {
                    SmartPlaylistImpl.this.setName(name);
                }
                public int getUniqueID() {
                    return SmartPlaylistImpl.this.uniqueid;
                }
                public int getRevision() {
                    return revision;
                }
                public Playlist sort(String sort, String filter) {
                    return SmartPlaylistImpl.this.sort(sort, filter);
                }
                public String toString() {
                     return SmartPlaylistImpl.this.toString()+"("+sort+","+filter+")";
                }
                public boolean equals(Object o) {
                     return o == this;
                }
            });
        }
        return pl;
    }

    public int size() {
        return getTracks().size();
    }

}
