package org.qtunes.ff.spi.mp3;

import org.qtunes.core.*;
import org.qtunes.db.*;
import java.util.*;
import java.io.*;
import java.nio.*;

/**
 * A class to map Tags (actual values) to Fields (logical values).
 */
class TagMapper {
    
    private static Map<String,Field> tagmap = new HashMap<String,Field>();
    private final MP3Handler handler;

    static {
        tagmap.put("TIT1", Field.Group);
        tagmap.put("TT1", Field.Group);
        tagmap.put("TPE1", Field.Artist);
        tagmap.put("TP1", Field.Artist);
        tagmap.put("TPE2", Field.AlbumArtist);
        tagmap.put("TP2", Field.AlbumArtist);
        tagmap.put("TPE3", Field.Conductor);
        tagmap.put("TP3", Field.Conductor);
        tagmap.put("TALB", Field.Album);
        tagmap.put("TAL", Field.Album);
        tagmap.put("TIT2", Field.Title);
        tagmap.put("TT2", Field.Title);
        tagmap.put("TCMP", Field.Compilation);
        tagmap.put("TCP", Field.Compilation);
        tagmap.put("TCOM", Field.Composer);
        tagmap.put("TCM", Field.Composer);
        tagmap.put("TENC", Field.Encoder);
        tagmap.put("TEN", Field.Encoder);
        tagmap.put("TSSE", Field.EncoderSettings);
        tagmap.put("TSS", Field.EncoderSettings);
        tagmap.put("TBPM", Field.BPM);
        tagmap.put("TBP", Field.BPM);
        tagmap.put("TPUB", Field.Publisher);
        tagmap.put("TPB", Field.Publisher);
        tagmap.put("TLAN", Field.Language);
        tagmap.put("TLA", Field.Language);
        tagmap.put("TMED", Field.Media);
        tagmap.put("TCOP", Field.Copyright);
        tagmap.put("TCR", Field.Copyright);
        tagmap.put("TOPE", Field.OriginalArtist);
        tagmap.put("TOA", Field.OriginalArtist);
        tagmap.put("TOAL", Field.OriginalAlbum);
        tagmap.put("TOT", Field.OriginalAlbum);
        tagmap.put("TEXT", Field.Lyricist);
        tagmap.put("TXT", Field.Lyricist);
        tagmap.put("WXXX", Field.URL);
        tagmap.put("WOAF", Field.URL);
        tagmap.put("WOAR", Field.URL);
        tagmap.put("WOAS", Field.URL);
    }

    TagMapper(MP3Handler handler) {
        this.handler = handler;
    }

    static boolean hasMap(String tag) {
        return tagmap.containsKey(tag);
    }

    //----------------------------------------------------------------------
    // Reading
    //----------------------------------------------------------------------

    void setFields(Tag tag, FieldMap fields) throws IOException {
        String name = tag.getName();
        boolean invalid = tag.isInvalid();
        if ("COMM".equals(name) || "COM".equals(name)) {
            String type = ((TagComment)tag).getType();
            String val = ((TagComment)tag).getValue();
            if (type.equals("iTunes_CDDB_1")) {
                if (!fields.containsKey(Field.CDDB) || name.length() > 3) {
                    fields.put(Field.CDDB, val);
                }
            } else if (type.equals("iTunes_CDDB_IDs")) {
                if (!fields.containsKey(Field.CDDB) || name.length() > 3) {
                    fields.put(Field.CDDB, val);
                }
            } else if (type.equals("iTunPGAP") && val.equals("1")) {
                if (!fields.containsKey(Field.Gapless) || name.length() > 3) {
                    fields.put(Field.Gapless, true);
                }
            } else if (type.equals("iTunNORM")) {
                try {
                    String[] com = val.trim().split(" ");
                    float lval1 = Integer.parseInt(com[0], 16) / 1000f;
                    float rval1 = Integer.parseInt(com[1], 16) / 1000f;
                    float lval2 = Integer.parseInt(com[2], 16) / 2500f;
                    float rval2 = Integer.parseInt(com[3], 16) / 2500f;
                    float gain = (float)(-10 * Math.log(Math.max(Math.max(lval1, lval2), Math.max(rval1, rval2))) / Math.log(10));
                    if (gain >= -25 && gain <= 25 && (!fields.containsKey(Field.Gain) || name.length() > 3)) {
                        fields.put(Field.Gain, new Float(gain));
                    }
                } catch (Exception e) {
                    e.printStackTrace();
                }
            } else if (!badComment(type, val)) {
                String oldcomment = (String)fields.get(Field.Comment);
                if (oldcomment == null || oldcomment.length() <= 2) {
                    fields.put(Field.Comment, val);
                }
            }
        } else if ("TYER".equals(name) || "TYE".equals(name)) {
            int val = (int)((TagNumber)tag).getValue();
            if (val >= 1800 && (!fields.containsKey(Field.Year) || name.length() > 3)) {
                fields.put(Field.Year, Integer.valueOf(val));
            }
        } else if ("TDRC".equals(name)) {
            fields.put(Field.Year, Integer.valueOf(((TagTimestamp)tag).getCalendar().get(Calendar.YEAR)));
        } else if ("TCP".equals(name) || "TCMP".equals(name)) {                 // non-standard iTunes
            if (!fields.containsKey(Field.Compilation) || name.length() > 3) {  // ID3v23/4 overrides v22
                if (((TagText)tag).getValue().equals("1")) {
                    fields.put(Field.Compilation, true);
                } else {
                    fields.remove(Field.Compilation);
                }
            }
        } else if ("TSRC".equals(name)) {
            String val = ((TagText)tag).getValue();
            if (val.length() == 12) {
                fields.put(Field.ISRC, val);
            }
        } else if ("TBPM".equals(name) || "TBP".equals(name)) {
            try {
                float val = Float.parseFloat(((TagText)tag).getValue());
                if (val > 0 && (!fields.containsKey(Field.BPM) || name.length() > 3)) {
                    fields.put(Field.BPM, val);
                }
            } catch (Exception e) { }
        } else if ("TCON".equals(name) || "TCO".equals(name)) {
            String val = ((TagGenre)tag).getValue();
            if (val.length() > 0 && (!fields.containsKey(Field.Genre) || name.length() > 3)) {
                fields.put(Field.Genre, val);
            }
        } else if (name.equals("TRCK") || name.equals("TRK")) {
            int pos = ((TagPosition)tag).getPosition();
            int max = ((TagPosition)tag).getRange();
            if (pos != 0 && (!fields.containsKey(Field.Track) || name.length() > 3)) {
                fields.put(Field.Track, Integer.valueOf(pos));
            }
            if (max != 0 && (!fields.containsKey(Field.NumTracks) || name.length() > 3)) {
                fields.put(Field.NumTracks, Integer.valueOf(max));
            }
        } else if (name.equals("TPOS") || name.equals("TPA")) {
            int pos = ((TagPosition)tag).getPosition();
            int max = ((TagPosition)tag).getRange();
            if (pos != 0 && (!fields.containsKey(Field.Disc) || name.length() > 3)) {
                fields.put(Field.Disc, Integer.valueOf(pos));
            }
            if (max != 0 && (!fields.containsKey(Field.NumDiscs) || name.length() > 3)) {
                fields.put(Field.NumDiscs, Integer.valueOf(max));
            }
        } else if (name.equals("APIC") || name.equals("PIC")) {
            int imagelen = ((TagImage)tag).getLength();
            int type = ((TagImage)tag).getType();
            String mimetype = ((TagImage)tag).getMIMEType();
            if (imagelen > 200 && (type==0 || type==3 || type==112) && mimetype.startsWith("image/") && (!fields.containsKey(Field.ImageLength) || name.length() > 3)) {
                long imageoff = ((TagImage)tag).getOffset();
                String description = ((TagImage)tag).getDescription();
                fields.put(Field.ImageMIMEType, mimetype);
                fields.put(Field.ImageOffset, Integer.valueOf((int)imageoff));
                fields.put(Field.ImageLength, Integer.valueOf(imagelen));
                if (description.length() > 0) {
                    fields.put(Field.ImageDescription, description);
                }
            }
        } else if (name.equals("MCDI")) {
            ByteBuffer data = tag.getContent();
            data.rewind();
            StringBuilder s = new StringBuilder(data.capacity() * 2);
            while (data.remaining() > 0) {
                int b = data.get() & 0xFF;
                if (b < 0x10) {
                    s.append('0');
                }
                s.append(Integer.toHexString(b));
            }
            fields.put(Field.MCDI, s.toString().replaceAll("0*$", ""));
        } else if (name.equals("RVAD")) {
            ByteBuffer data = tag.getContent();
            data.rewind();
            int sign = data.get() & 0xFF;
            int bits = data.get() & 0xFF;
            if (bits==16) {
                int igain1 = (((data.get()&0xFF)<<8) + (data.get()&0xFF)) * (((sign & 2)==0) ? -1 : 1);
                int igain2 = (((data.get()&0xFF)<<8) + (data.get()&0xFF)) * (((sign & 1)==0) ? -1 : 1);
                float gain = (float)Math.max(-96, Math.min(6, 20 * Math.log(((Math.max(igain1, igain2) / 256.0) + 255) / 255) / Math.log(10)));
                if (gain >= -25 && gain <= 25 && !fields.containsKey(Field.Gain)) {     // Don't replace, as we're not sure this is right
                    fields.put(Field.Gain, new Float(gain));
                }
            }
        } else if (name.equals("RVA2")) {
            ByteBuffer data = tag.getContent();
            data.rewind();
            int type = data.get() & 0xFF;
            if (type==1) {      // Master Volume
                int igain = ((data.get()&0xFF)<<8) + (data.get()&0xFF);
                float gain = igain / 512f;
                if (gain >= -25 && gain <= 25) {
                    fields.put(Field.Gain, new Float(gain));
                }
            }
        } else {
            Field<?> field = (Field)tagmap.get(name);
            if (field != null) {
                if (!fields.containsKey(field) || name.length() > 3) {
                    fields.putUntyped(field, ((TagText)tag).getValue());
                }
            }
        }
        if (invalid) {
            fields.put(Field.Warning, Boolean.TRUE);
        }
    }

    private static boolean badComment(String type, String val) {
        if (type.startsWith("iTun")) return true;
        if (type.equals("AudioGenie")) return true;
        if (type.equals("MusicMatch_Serial")) return true;
        if (val.startsWith("Encoded with Easy CD-DA")) return true;
        if (type.equals("ID3v1 Comment")) {
            if (val.startsWith(" 0000")) return true;               // AudioGenie in v1
            if (val.length() > 2 && val.charAt(2)=='+' && Character.isDigit(val.charAt(0)) && Character.isDigit(val.charAt(1))) return true;
            if (val.startsWith(" YEAR: ")) return true;
        }
        if (val.trim().length()==0) return true;
        return false;
    }

    //----------------------------------------------------------------------
    // Writing
    //----------------------------------------------------------------------

    /**
     * Search the tag list and return the tag that matches the specified
     * version and that has one of the specified names. Any other tags
     * with matching names are removed.
     */
    private static Tag findTag(List<Tag> tags, int version, String[] names) {
        Tag tag = null;
        for (int i=0;i<tags.size();i++) {
            Tag ttag = tags.get(i);
            String name = ttag.getName();
            for (int j=0;j<names.length;j++) {
                if (name.equals(names[j])) {
                    if (tag == null && j==(version==2 ? 0 : version==3 ? 1 : names.length-1)) {
                        tag = ttag;
                    } else {
                        tags.remove(i--);
                    }
                }
            }
        }
        return tag;
    }

    /**
     * Search the tag list and return the comment tag that matches the specified
     * comment type. Any subsequent duplicates are removed
     */
    private static TagComment findComment(List<Tag> tags, int version, String type) {
        TagComment tag = null;
        for (int i=0;i<tags.size();i++) {
            Tag ttag = tags.get(i);
            try {
                if (ttag instanceof TagComment && ((TagComment)ttag).getType().equals(type)) {
                    if (tag == null) {
                        tag = (TagComment)ttag;
                    } else {
                        tags.remove(i--);
                    }
                }
            } catch (IOException e) {}
        }
        return tag;
    }

    void setTags(Field field, Object value, List<Tag> tags, int version) {
        Tag newtag = null;
        if (field==Field.Track || field==Field.NumTracks || field==Field.Disc || field==Field.NumDiscs) {
            boolean istrack = field==Field.Track || field==Field.NumTracks;
            boolean isrange = field==Field.NumTracks || field==Field.NumDiscs;
            TagPosition oldtag = (TagPosition)findTag(tags, version, istrack ? new String[] { "TRK", "TRCK" } : new String[] { "TPA", "TPOS" });
            int position = 0, range = 0;
            try {
                if (isrange) {
                    position = oldtag==null ? 0 : oldtag.getPosition();
                    range = value==null ? 0 : (Integer)value;
                } else {
                    position = value==null ? 0 : (Integer)value;
                    range = oldtag==null ? 0 : oldtag.getRange();
                }
            } catch (IOException e) {}
            if (position > 0) {
                if (range < position) {
                    range = 0;
                }
                String name = istrack ? version==2 ? "TRK" : "TRCK" : version==2 ? "TPA" : "TPOS";
                newtag = new TagPosition(name, version, 0, position, range);
            }
            replaceTag(tags, oldtag, newtag);
        } else if (field==Field.Genre) {
            Tag oldtag = findTag(tags, version, new String[] { "TCO", "TCON" });
            if (value != null) {
                newtag = new TagGenre(version==2 ? "TCO" : "TCON", version, 0, (String)value);
            }
            replaceTag(tags, oldtag, newtag);
        } else if (field==Field.Gapless) {
            Tag oldtag = findComment(tags, version, "iTunPGAP");
            if (value != null) {
                newtag = new TagComment(version, 0, "eng", "iTunPGAP", "1");
            }
            replaceTag(tags, oldtag, newtag);
        } else if (field==Field.Compilation) {
            Tag oldtag = findTag(tags, version, new String[] { "TCP", "TCMP" });
            if (value != null) {
                newtag = new TagText(version==2 ? "TCP" : "TCMP", version, 0, "1");
            }
            replaceTag(tags, oldtag, newtag);
        } else if (field==Field.CDDB) {
            tags.remove(findComment(tags, version, "iTunes_CDDB_1"));
            Tag oldtag = findComment(tags, version, "iTunes_CDDB_IDs");
            if (value != null) {
                newtag = new TagComment(version, 0, "eng", "iTunes_CDDB_IDs", (String)value);
            }
            replaceTag(tags, oldtag, newtag);
        } else if (field==Field.Year) {
            Tag oldtag = findTag(tags, version, new String[] { "TYE", "TYER", "TDRC" });
            if (value != null) {
                newtag = new TagText(version==2 ? "TYE" : version==3 ? "TYER" : "TDRC", version, 0, value.toString());
            }
            replaceTag(tags, oldtag, newtag);
        } else if (field==Field.Comment) {
            Tag oldtag = null;
            for (int j=0;j<tags.size();j++) {
                Tag tag = tags.get(j);
                if (tag instanceof TagComment) {
                    try {
                        String type = ((TagComment)tag).getType();
                        String oldvalue = ((TagComment)tag).getValue();
                        if (!badComment(type, oldvalue)) {
                            if (oldtag == null) {
                                oldtag = tag;
                            } else {
                                tags.remove(j--);
                            }
                        }
                    } catch (IOException e) {
                        tags.remove(j--);
                    }
                }
            }
            if (value != null) {
                newtag = new TagComment(version, 0, "eng", "", (String)value);
            }
            replaceTag(tags, oldtag, newtag);
        } else if (field == Field.Gain) {
            Tag oldrgad = findTag(tags, version, new String[] { "RGAD", "RGAD" });
            Tag oldrva2 = findTag(tags, version, new String[] { null, "RVAD", "RVA2" });
            TagComment oldnorm = findComment(tags, version, "iTunNORM");
            Tag newnorm = null, newrva2 = null, newrgad = null;

            if (value != null) {
                float gain = ((Number)value).floatValue();
                if (handler.getGainStorage("itunes")) {
                    String[] norm = new String[10];
                    if (oldnorm != null) {
                        try {
                            norm = oldnorm.getValue().trim().split(" ");
                        } catch (Exception e) { }
                    }
                    norm[0] = norm[1] = "00000000" + Long.toHexString(Math.min(65534, Math.round(Math.pow(10, gain / -10) * 1000))).toUpperCase();
                    norm[2] = norm[3] = "00000000" + Long.toHexString(Math.min(65534, Math.round(Math.pow(10, gain / -10) * 2500))).toUpperCase();
                    StringBuffer sb = new StringBuffer();
                    for (int i=0;i<10;i++) {
                        sb.append(" ");
                        sb.append(norm[i]==null ? "00000000" : norm[i].substring(norm[i].length()-8));
                    }
                    newnorm = new TagComment(version, 0, "eng", "iTunNORM", sb.toString());
                }

                if (handler.getGainStorage("replaygain")) {
                    // ReplayGain tag. Base level is moved from 89db to 83db, and we don't both
                    // calculating the peak amplitude, setting it to 0xFFFF
                    int rpgn = 0x2C00 + (gain < 0 ? 0x200 : 0) + Math.max(0x1FF, Math.round(Math.abs(gain-6)*10));
                    byte[] rpgndata = new byte[] { -1, -1, -1, -1, (byte)(rpgn>>8), (byte)rpgn, 0, 0 };
                    newrgad = new Tag("RGAD", version, 0, rpgndata);
                }

                if (handler.getGainStorage("rva")) {
                    if (version == 24) {
                        short igain = (short)(gain * 512);
                        byte[] data = new byte[] { 0x00, (byte)((igain>>8)&0xFF), (byte)(igain&0xFF), 0, 0 };
                        newrva2 = new Tag("RVA2", version, 0, data);
                    } else {
                        // Spec is useless on what the value actually means, so this lifted from From SlimServer 6.51/MP3::Info.pm
                        if (gain >= -96 && gain <= 6) {
                            int igain = (int)(((Math.pow(10, gain / 20) * 255) - 255) * 256);
                            int peak = 0;

                            byte ig1 = (byte)((igain>>8)&0xFF);
                            byte ig2 = (byte)(igain&0xFF);
                            byte pk1 = (byte)((peak>>8)&0xFF);
                            byte pk2 = (byte)(peak&0xFF);
                            byte[] data = new byte[] { (byte)(igain >=0 ? 0x11 : 0x00), (byte)0x10, ig1, ig2, ig1, ig2, pk1, pk2, pk1, pk2 };
                            newrva2 = new Tag("RVAD", version, 0, data);
                        }
                    }
                }
            }
            replaceTag(tags, oldnorm, newnorm);
            replaceTag(tags, oldrva2, newrva2);
            replaceTag(tags, oldrgad, newrgad);

        } else {
            for (Iterator<Map.Entry<String,Field>> i = tagmap.entrySet().iterator();i.hasNext();) {
                Map.Entry<String,Field> e = i.next();
                String name = e.getKey();
                if (e.getValue() == field && name.length() == (version == 2 ? 3 : 4)) {
                    Tag oldtag = null;
                    for (int j=0;j<tags.size();j++) {
                        Tag tag = tags.get(j);
                        if (tagmap.get(tag.getName()) == field) {
                            if (oldtag == null) {
                                oldtag = tag;
                            } else {
                                tags.remove(j--);
                            }
                        }
                    }
                    if (value != null) {
                        newtag = new TagText(name, version, 0, value.toString());
                    }
                    replaceTag(tags, oldtag, newtag);
                }
            }
        }
    }

    private void replaceTag(List<Tag> tags, Tag oldtag, Tag newtag) {
        int ix = oldtag == null ? -1 : tags.indexOf(oldtag);
        if (ix >= 0) {
            if (newtag != null) {
                if (!newtag.equals(oldtag)) {
                    tags.set(ix, newtag);
                }
            } else {
                tags.remove(ix);
            }
        } else if (newtag != null) {
            tags.add(newtag);
        }
    }

}
