package org.qtunes.player.spi;

import org.qtunes.player.*;
import org.qtunes.core.*;
import org.qtunes.auth.*;
import org.qtunes.db.*;
import java.util.*;
import java.io.*;

/**
 * A Simple Playlist that has a manually managed list of tracks
 */
public class PlayerPlaylist implements ManualPlaylist {

    private final PlayerImpl player;
    private List<TrackSession> list = new ArrayList<TrackSession>();
    private List<TrackSession> shuffle, newtracks;
    private boolean wasempty;

    PlayerPlaylist(PlayerImpl player) {
        this.player = player;
    }

    public String getName() {
        throw new UnsupportedOperationException();
    }

    public void setName(String name) {
        throw new UnsupportedOperationException();
    }

    public int getUniqueID() {
        return 0;
    }

    public int getRevision() {
        return player.getRevision();
    }

    /**
     * Add the tracks specified in this Playlist
     * @param playlist the tracks to add
     */
    public void addAll(Playlist playlist) {
        synchronized(player) {
            addAll(playlist, size());
        }
    }

    /**
     * Add the specified tracks to this Playlist, but only if they're not duplicates.
     * @param playlist the tracks to add
     */
    public void mergeAll(Playlist playlist) {
        synchronized(player) {
            mergeAll(playlist, size());
        }
    }

    /**
     * Add the specified tracks to this Playlist at the specified point
     * @param playlist the tracks to add
     * @param index the insertion position in the list
     */
    public void addAll(Playlist playlist, int index) {
        synchronized(player) {
            begin();
            Session session = Session.getCurrentSession();
            List<Track> newtracks = playlist.getTracks();
            for (int i=0;i<newtracks.size();i++) {
                Track track = newtracks.get(i);
                addTS(index++, new TrackSession(track, session));
            }
            end();
        }
    }

    /**
     * Add the specified tracks to this Playlist at the specified point,
     * but only if they're not duplicates
     * @param playlist the tracks to add
     * @param index the insertion position in the list
     */
    public void mergeAll(Playlist playlist, int index) {
        synchronized(player) {
            begin();
            List<Track> oldtracks = getTracks();
            List<Track> newtracks = playlist.getTracks();
            newtracks.removeAll(oldtracks);
            Session session = Session.getCurrentSession();
            for (int i=0;i<newtracks.size();i++) {
                Track track = newtracks.get(i);
                addTS(index++, new TrackSession(track, session));
            }
            end();
        }
    }

    /**
     * Add the specified Track to the Playlist.
     * @param track the Track to add
     */
    public void add(Track track) {
        synchronized(player) {
            begin();
            add(track, size());
            end();
        }
    }

    /**
     * Add the specified Track to the Playlist.
     * @param track the Track to add
     * @param index the insertion position in the list
     */
    public void add(Track track, int index) {
        synchronized(player) {
            begin();
            addTS(index, new TrackSession(track, Session.getCurrentSession()));
            end();
        }
    }

    public void clear() {
        synchronized(player) {
            begin();
            list.clear();
            if (shuffle!=null) {
                shuffle.clear();
            }
            end();
        }
    }

    public Track remove(int index) {
        synchronized(player) {
            begin();
            TrackSession ts = doRemove(index);
            end();
            return ts.track;
        }
    }

    public boolean removeTrack(Track track) {
        synchronized(player) {
            begin();
            boolean found = false;
            for (int i=0;i<list.size();i++) {
                if (list.get(i).track==track) {
                    doRemove(i--);
                    found = true;
                }
            }
            if (found) {
                end();
            }
            return found;
        }
    }

    public void removeAll(Playlist playlist) {
        synchronized(player) {
            Collection<Track> newtracks = new HashSet<Track>(playlist.getTracks());
            boolean found = false;
            begin();
            for (int i=0;i<list.size();i++) {
                if (newtracks.contains(list.get(i).track)) {
                    doRemove(i--);
                    found = true;
                }
            }
            if (found) {
                end();
            }
        }
    }

    public void removeAll(BitSet bitset) {
        synchronized(player) {
            if (bitset.cardinality() > 0) {
                begin();
                int diff = 0;
                for (int i=bitset.nextSetBit(0);i>=0;i=bitset.nextSetBit(i+1)) {
                    TrackSession ts = list.remove(i - (diff++));
                    if (shuffle!=null) {
                        shuffle.remove(ts);
                    }
                }
                end();
            }
        }
    }

    public List<Track> getTracks() {
        synchronized(player) {
            List<Track> out = new ArrayList<Track>(list.size());
            for (int i=0;i<list.size();i++) {
                out.add(list.get(i).track);
            }
            return out;
        }
    }

    public Playlist sort(String key, String filter) {
        return this;
    }

    public int size() {
        synchronized(player) {
            return list.size();
        }
    }

    /**
     * Get the next TrackSession or null if the playlist has ended
     * @param ts the current TrackSession
     * @param direction 1 or -1, for forward or backward
     * @param auto true if the track is automatically moving, false if it's a click
     */
    TrackSession getNextTrackSession(TrackSession ts, int direction, boolean auto, int repeat) {
        synchronized(player) {
            TrackSession oldts = ts;
            if (auto && repeat==1) {
                ts = list.contains(ts) ? ts : null;
            } else {
                List<TrackSession> l = shuffle==null ? list : shuffle;
                int index = l.indexOf(ts);
                if (index < 0) {
                    ts = null;
                } else {
                    index += direction;
                    if (repeat==2) {
                        index = index % list.size();
                    }
                    ts = index < 0 || index >= l.size() ? null : l.get(index);
                }
            }
            return ts;
        }
    }

    TrackSession getTrackSession(int index) {
        synchronized(player) {
            return index < 0 || index >= list.size() ? null : shuffle==null ? list.get(index) : shuffle.get(index);
        }
    }

    //--------------------------------------------------------------

    private void begin() {
        if (shuffle!=null) {
            newtracks = new ArrayList<TrackSession>();
        }
        wasempty = list.isEmpty();
    }

    private void end() {
        if (shuffle!=null) {
            Collections.shuffle(newtracks);
            shuffle.addAll(newtracks);
            newtracks = null;
        }
        player.notifyListChanged(wasempty);
    }

    private void addTS(int ix, TrackSession ts) {
        if (ix==size()) {
            list.add(ts);
        } else {
            list.add(ix, ts);
        }
        if (shuffle!=null) {
            newtracks.add(ts);
        }
    }

    private TrackSession doRemove(int index) {
        TrackSession ts = list.remove(index);
        if (shuffle!=null) {
            shuffle.remove(ts);
        }
        return ts;
    }

    int indexOf(TrackSession session) {
        synchronized(player) {
            return list.indexOf(session);
        }
    }

    void setShuffle(boolean shuffle) {
        synchronized(player) {
            if (shuffle) {
                this.shuffle = new ArrayList<TrackSession>(list);
                Collections.shuffle(this.shuffle);
            } else {
                this.shuffle = null;
            }
        }
    }

    boolean isShuffle() {
        synchronized(player) {
            return shuffle != null;
        }
    }

}
