package org.qtunes.speaker.spi.javasound;

import org.qtunes.core.*;
import org.qtunes.speaker.*;
import org.qtunes.player.Player;
import javax.sound.sampled.*;
import java.io.*;
import java.io.*;
import java.util.*;

public class SpeakerImpl implements Speaker {

    private ServiceContext context;
    private SourceDataLine line;
    private Mixer.Info mixer;
    private Player player;
    private volatile Exception error; 
    private int volumeadjust;
    private boolean started;

    private synchronized void setError(Exception e) {
        this.error = e;
        if (error != null) {
            context.warn("Speaker Error", e);
            context.fireEvent("speakerError", new Object[] { "exception", this.error });
        }
    }

    public Exception getError() {
        return error;
    }

    public synchronized Player getPlayer() {
        return player;
    }

    public synchronized void setPlayer(Player player) {
        if (player != this.player) {
            if (this.player != null) {
                this.player.removeSpeaker(this);
            }
            this.player = player;
            if (player != null) {
                player.addSpeaker(this);
            }
        }
    }

    public String getDisplayName() {
        return context.getProperty("display");
    }

    public int getUniqueId() {
        try {
            return Integer.parseInt(context.getProperty("uid"));
        } catch (Exception e) {
            return getDisplayName().hashCode() & 0x7FFFFFFF;
        }
    }

    public synchronized boolean isAvailable() {
        return line != null;
    }

    public synchronized boolean hasGain() {
        return line != null && line.isControlSupported(FloatControl.Type.MASTER_GAIN);
    }

    public synchronized void setMixer(Mixer.Info mixer) {
        try {
            line = AudioSystem.getSourceDataLine(new AudioFormat(44100, 16, 2, true, false), mixer);
            line = null;
            this.mixer = mixer;
            context.putProperty("mixer", mixer.getName());
        } catch (Exception e) {
            setError(e);
        }
    }

    public void setMixer(String name) {
        Mixer.Info[] mixers = AudioSystem.getMixerInfo();
        for (int i=0;i<mixers.length;i++) {
            if (mixers[i].getName().equals(name)) {
                setMixer(mixers[i]);
            }
        }
    }

    public synchronized Mixer.Info getMixer() {
        return line != null ? mixer : null;
    }

    public int getDelay() {
        return 0;
    }

    public int getBufferSize() {
        return 1408;
    }

    public synchronized void write(byte[] buf, int off, int len) {
        if (error == null) {
            try {
                if (line == null) {
                    throw new IllegalStateException("Not connected");
                }
                if (!started) {
                    line.start();
                    started = true;
                    context.fireEvent("speakerStarted", null);
                }
                line.write(buf, off, len);
            } catch (Exception e) {
                setError(e);
            }
        }
    }

    public synchronized void flush() {
        if (error == null) {
            started = false;
            try {
                if (line == null) {
                    throw new IOException("Not connected");
                }
                line.stop();
                line.flush();
                context.fireEvent("speakerStopped", null);
            } catch (Exception e) {
                setError(e);
            }
        }
    }

    public synchronized void drain() {
        if (error == null) {
            started = false;
            try {
                if (line == null) {
                    throw new IOException("Not connected");
                }
                line.drain();
                context.fireEvent("speakerStopped", null);
            } catch (Exception e) {
                setError(e);
            }
        }
    }

    public boolean isOpen() {
        return line != null;
    }

    public synchronized void close() {
        started = false;
        if (line!=null) {
            try {
                line.stop();
                line.close();
                line = null;
            } catch (Exception e) { }
        }
    }


    public synchronized void open(float sampleRate, int sampleSizeInBits, int channels, boolean signed, boolean bigEndian) {
        AudioFormat format = new AudioFormat(sampleRate, sampleSizeInBits, channels, signed, bigEndian);
        if (line != null && line.isOpen() && !format.equals(line.getFormat())) {
            close();
        }
        try {
            if (line == null) {
                line = AudioSystem.getSourceDataLine(format, mixer);
                if (line == null) {
                    throw new IllegalStateException("Can't get Line for "+format+" from "+mixer);
                }
            }
            error = null;
            line.open(format);
            line.start();
            started = true;
            context.fireEvent("speakerOpened", null);
        } catch (Exception e) {
            setError(e);
        }
    }

    public synchronized void setVolumeAdjustment(int volumeadjust) {
        this.volumeadjust = volumeadjust;
        context.putProperty("volumeadjust", Integer.toString(volumeadjust));
    }

    public synchronized int getVolumeAdjustment() {
        return volumeadjust;
    }

    public synchronized void setGain(float gain) {
        if (hasGain()) {
            if (gain!=gain) {
                gain = -100;
            }
            try {
                if (line == null) {
                    throw new IOException("Not connected");
                }
                FloatControl control = (FloatControl)line.getControl(FloatControl.Type.MASTER_GAIN);
                control.setValue(gain);
            } catch (Exception e) {
                setError(e);
            }
        }
    }

    //----------------------------------------------------------------------------

    public void startService(ServiceContext context) {
        this.context = context;
        if (getDisplayName()==null) {
            throw new IllegalStateException("No \"display\" property");
        }
        if (context.getProperty("mixer") != null) {
            setMixer(context.getProperty("mixer"));
        }
        try {
            volumeadjust = Integer.parseInt(context.getProperty("volumeadjust"));
        } catch (Exception e) { }

        if (mixer == null) {
            Mixer.Info[] mixers = AudioSystem.getMixerInfo();
            if (mixers != null && mixers.length > 0) {
                setMixer(mixers[0]);
            } else {
	        context.error("No JavaSound mixers", null);
		context.stop();
		context.removeService();
	    }
        }
    }

    public void stopService(ServiceContext context) {
    }

    public ServiceContext getContext() {
        return context;
    }

    public Map<String,Object> reportState() {
        Map<String,Object> map = new LinkedHashMap<String,Object>();
        map.put("name", getDisplayName());
        map.put("volumeadjust", getVolumeAdjustment());
        map.put("hasgain", hasGain());
        return map;
    }

}
