package org.qtunes.tagger;

import java.io.*;
import java.nio.*;
import java.nio.channels.*;
import java.util.*;
import java.text.DecimalFormat;
import org.qtunes.replaygain.*;
import org.qtunes.db.*;
import org.qtunes.ff.*;
import org.qtunes.ff.spi.mp3.MP3Handler;

class NormalizerTool implements Action {
    private static final int AUTO = 4;
    private static final int REMOVE = 3;
    private static final int ALBUM = 2;
    private static final int TRACK = 1;
    private static final int NONE = 0;
    private static final float THRESHOLD = 0.15f;
    private PrintStream out = System.out;
    private Normalizer normalizer;
    private float gain = 90;
    private int mode, debug;
    private boolean dryrun;
    private Map<Object,List<File>> filelist;

    NormalizerTool() {
        normalizer = new Normalizer();
        filelist = new LinkedHashMap<Object,List<File>>();
    }

    public void setOutput(PrintStream out) {
        this.out = out;
        normalizer.setDebug(out, debug > 0);
    }

    public void setDryRun(boolean on) {
        dryrun = on;
    }

    public void bumpDebug() {
        debug ++;
        normalizer.setDebug(out, debug > 0);
    }

    void setMode(String mode, float gain) {
        this.gain = gain;
        if ("track".equals(mode)) {
            this.mode = TRACK;
        } else if ("album".equals(mode)) {
            this.mode = ALBUM;
        } else if ("auto".equals(mode)) {
            this.mode = AUTO;
        } else if ("remove".equals(mode)) {
            this.mode = REMOVE;
        } else {
            throw new IllegalArgumentException(mode);
        }
    }

    void setTagType(String mode) {
        MP3Handler reader = (MP3Handler)TrackReader.getTrackReader("mp3");
        reader.setGainStorage("itunes", mode.contains("itunes"));
        reader.setGainStorage("replaygain", mode.contains("replaygain"));
        reader.setGainStorage("rva", mode.contains("rva"));
        if (!reader.getGainStorage("itunes") && !reader.getGainStorage("replaygain") && !reader.getGainStorage("rva")) {
            throw new IllegalArgumentException("No Gain tag specified - must be comma seperated list of [itunes,replaygain,rva]");
        }
    }

    void setDatabase(File filename) {
        normalizer.setDatabase(filename);
    }

    public void process(File file) throws Exception {
        if (mode == REMOVE) {
            normalizer.remove(file);
        } else {
            normalizer.enqueue(file);

            Object key;
            if (mode == ALBUM) {
                key = null;
            } else if (mode == TRACK) {
                key = file;
            } else if (mode == AUTO) {
                TrackReader reader = TrackReader.getTrackReader(file);
                FieldMap map = reader.read(file);
                if (map.get(Field.Gapless) == Boolean.TRUE) {
                    key = map.get(Field.Album) + map.get(Field.Disc);
                } else {
                    key = file;
                }
            } else {
                return;
            }
            if (key == file) {
                filelist.put(key, Collections.singletonList(file));
            } else {
                List<File> l = filelist.get(key);
                if (l == null) {
                    filelist.put(key, l = new ArrayList<File>());
                }
                l.add(file);
            }
        }
    }

    public void close() throws Exception {
        while (!normalizer.waitForCompletion()) {}
        normalizer.shutdown();
        normalizer.saveDatabase();

        DecimalFormat df = new DecimalFormat("#0.0");
        for (Iterator<List<File>> i = filelist.values().iterator();i.hasNext();) {
            List<File> list = i.next();
            List<GainAnalysis> analyses = new ArrayList<GainAnalysis>(list.size());
            for (int j=0;j<list.size();j++) {
                File file = list.get(j);
                GainAnalysis analysis = normalizer.getGain(file);
                if (analysis != null) {
                    analyses.add(analysis);
                } else {
                    if (list.size() == 1 || filelist.size() == 1) {
                        System.out.println("# Analysis failed on \""+file+"\" - skipping");
                    } else {
                        System.out.println("# Analysis failed on \""+file+"\" - skipping the following "+list.size()+" tracks:");
                        for (int k=0;k<list.size();k++) {
                            System.out.println("## "+list.get(k));
                        }
                    }
                    list = null;
                    break;
                }
            }

            if (list != null) {
                float mod = GainAnalysis.getAdjustment(analyses, this.gain);
                for (int j=0;j<list.size();j++) {
                    File file = list.get(j);
                    TrackReader reader = TrackReader.getTrackReader(file);
                    if (reader instanceof TrackWriter) {
                        FieldMap fields = reader.read(file);
                        Float z = fields.get(Field.Gain);
                        float oldmod = z==null ? 0 : z.floatValue();
                        float diff = Math.abs(oldmod - mod);
                        if (diff > THRESHOLD) {
                            if (debug > 0) {
                                System.out.println("debug: \""+file+"\" mod was "+df.format(oldmod)+"dB, now "+df.format(mod)+"dB");
                            }
                            fields.put(Field.Gain, Float.valueOf(mod));
                            if (!dryrun) {
                                ((TrackWriter)reader).write(file, fields);
                            }
                        } else if (debug > 1) {
                            System.out.println("debug: \""+file+"\" mod already set to "+df.format(oldmod)+"dB");
                        }
                    }
                }
            }
        }
    }

}
