package org.qtunes.tagger;

import org.qtunes.core.*;
import org.qtunes.core.util.*;
import org.qtunes.ff.*;
import org.qtunes.db.*;
import java.io.*;
import java.util.*;
import java.text.*;
import java.text.Normalizer;
import java.nio.*;
import java.nio.channels.*;
import javax.sound.sampled.*;
import javax.script.*;

class Scripter implements Action {

    private PrintStream out = System.out;
    private String function;
    private ScriptEngine engine;
    private int debug;
    private boolean dryrun, caseinsensitive, printrename;

    public void setOutput(PrintStream out) {
        this.out = out;
    }

    public void setCaseInsensitive(boolean caseinsensitive) {
        this.caseinsensitive = caseinsensitive;
    }

    public void setPrintRename(boolean print) {
        this.printrename = print;
    }

    public void setDryRun(boolean dryrun) {
        this.dryrun = dryrun;
    }

    public void bumpDebug() {
        this.debug++;
    }

    private Object javaToJS(Object key, FieldMap fieldmap, File file, Object o) {
        if (key.equals("imagedata")) {
            Number offset = fieldmap.get(Field.ImageOffset);
            Number length = fieldmap.get(Field.ImageLength);
            if (offset != null && length != null) {
                ByteBuffer buf = ByteBuffer.allocate(length.intValue());
                FileChannel in = null;
                try {
                    in = new FileInputStream(FileUtils.getAbsoluteFile(file)).getChannel();
                    in.position(offset.longValue());
                    in.read(buf);
                    return buf;
                } catch (IOException e) {
                    e.printStackTrace(out);
                } finally {
                    if (in != null) try { in.close(); } catch (IOException e) {}
                }
            }
            return "";
        }
        if (key instanceof Field && o == null) {
            Class type = ((Field)key).getType();
            if (type == String.class) {
                return "";
            } else if (type == Boolean.class) {
                return Boolean.FALSE;
            } else {
                return Integer.valueOf(0);
            }
        }
        return o;
    }

    private static Object cast(Field f, Object o) {
        if (o == null) {
            return o;
        } else if (f.getType() == String.class) {
            return o.toString();
        } else if (f.getType() == Integer.class) {
            if (o instanceof Boolean) {
                return Integer.valueOf(o == Boolean.TRUE ? 1 : 0);
            } else if (o instanceof Integer) {
                return o;
            } else if (o instanceof Number) {
                return Integer.valueOf(((Number)o).intValue());
            } else if (o instanceof String) {
                return Integer.valueOf((String)o);
            }
        } else if (f.getType() == Float.class) {
            if (o instanceof Float) {
                return o;
            } else if (o instanceof Number) {
                return Float.valueOf(((Number)o).floatValue());
            } else if (o instanceof String) {
                return Float.valueOf((String)o);
            }
        } else if (f.getType() == Long.class) {
            if (o instanceof Number) {
                return Long.valueOf(((Number)o).longValue());
            } else if (o instanceof String) {
                return Long.valueOf((String)o);
            }
        } else if (f.getType() == Boolean.class) {
            if (o instanceof Boolean ) {
                return o;
            } else if (o instanceof Number) {
                int v = ((Number)o).intValue();
                return v==0 ? Boolean.FALSE : Boolean.TRUE;
            } else if (o instanceof String) {
                return Boolean.valueOf((String)o);
            }
        }
        throw new ClassCastException("Can't convert "+o.getClass().getName()+" "+o+" to "+f.getType().getName()+" for '"+f.getName().toLowerCase()+"'");
    }

    void setFunction(String function) throws Exception {
        this.function = function;
        if (engine != null) {
            ((Invocable)engine).invokeFunction("before", new Object[] { function });
        }
    }

    void setScript(Reader reader) throws Exception {
        ScriptEngineManager manager = new ScriptEngineManager();
        engine = manager.getEngineByName("JavaScript");
        engine.setBindings(engine.createBindings(), ScriptContext.ENGINE_SCOPE);
        engine.getBindings(ScriptContext.ENGINE_SCOPE).put("__out", out);
        engine.getBindings(ScriptContext.ENGINE_SCOPE).put("debug", Integer.valueOf(debug));

        Reader initreader = new InputStreamReader(getClass().getResourceAsStream("init.js"), "UTF-8");
        CompiledScript script = ((Compilable)engine).compile(initreader);
        initreader.close();
        script.eval();

        script = ((Compilable)engine).compile(reader);
        reader.close();
        script.eval();

        if (function != null) {
            ((Invocable)engine).invokeFunction("before", new Object[] { function });
        }
    }

    @SuppressWarnings("unchecked")
    public void process(File file) throws Exception {
        TrackReader reader = TrackReader.getTrackReader(file);
        if (reader instanceof TrackWriter) {

            Track track = new Track(reader.read(file));
            ((Invocable)engine).invokeFunction("action", new Object[] { function, track });

            if (file != null) {
                boolean changed = track.isChanged();
                FieldMap map = track.getMap();
                if (!track.isTouched()) {
                    map.remove(Field.LastModified);
                }
                if (changed && (dryrun || ((TrackWriter)reader).write(file, map))) {
                    out.println("# Updated \""+file+"\"");
                } else {
                    changed = false;
                }
                File newfile = new File(track.getFile());
                if (!FileUtils.equals(newfile, file)) {
                    if (printrename) {
                        out.println("mv \""+file+"\" \""+newfile+"\"");
                    } else {
                        File absfile = FileUtils.getAbsoluteFile(file);
                        File absnewfile = FileUtils.getAbsoluteFile(newfile);
                        if (dryrun || ((absnewfile.getParentFile().exists() || absnewfile.getParentFile().mkdirs()) && absfile.renameTo(absnewfile))) {
                            out.println("# Renamed \""+file+"\" to \""+newfile+"\"");
                        } else {
                            out.println("# Failed renaming \""+absfile+"\" to \""+absnewfile+"\"");
                        }
                    }
                    changed = true;
                }
                if (!changed) {
                    if (debug > 0) out.println("# Skipped \""+file+"\"");
                }
            }
        }
    }

    public void close() throws Exception {
        ((Invocable)engine).invokeFunction("after", new Object[] { function });
    }

    public class Track {
        private final FieldMap map;
        private byte[] image;
        private boolean changed, touched;

        Track(FieldMap map) {
            this.map = map;
        }

        FieldMap getMap() {
            return map;
        }

        boolean isChanged() {
            return changed;
        }

        boolean isTouched() {
            return touched;
        }

        public String getTitle()                        { return getS(Field.Title); }
        public void setTitle(String value)              { put(Field.Title, value); }
        public String getArtist()                       { return getS(Field.Artist); }
        public void setArtist(String value)             { put(Field.Artist, value); }
        public String getAlbum()                        { return getS(Field.Album); }
        public void setAlbum(String value)              { put(Field.Album, value); }
        public String getAlbumartist()                  { return getS(Field.AlbumArtist); }
        public void setAlbumartist(String value)        { put(Field.AlbumArtist, value); }
        public String getGenre()                        { return getS(Field.Genre); }
        public void setGenre(String value)              { put(Field.Genre, value); }
        public int getTrack() 			        { return getI(Field.Track); }
        public void setTrack(int value) 		{ put(Field.Track, value); }
        public int getNumtracks() 			{ return getI(Field.NumTracks); }
        public void setNumtracks(int value) 		{ put(Field.NumTracks, value); }
        public Integer getDisc() 			{ return getI(Field.Disc); }
        public void setDisc(int value) 			{ put(Field.Disc, value); }
        public Integer getNumdiscs() 			{ return getI(Field.NumDiscs); }
        public void setNumdiscs(int value) 		{ put(Field.NumDiscs, value); }
        public String getComposer() 			{ return getS(Field.Composer); }
        public void setComposer(String value) 		{ put(Field.Composer, value); }
        public int getYear() 			        { return getI(Field.Year); }
        public void setYear(int value) 			{ put(Field.Year, value); }
        public boolean getCompilation() 		{ return getB(Field.Compilation); }
        public void setCompilation(boolean value) 	{ put(Field.Compilation, value); }
        public String getComment() 			{ return getS(Field.Comment); }
        public void setComment(String value) 		{ put(Field.Comment, value); }
        public String getCddb() 			{ return getS(Field.CDDB); }
        public void setCddb(String value) 		{ put(Field.CDDB, value); }
        public String getMcdi() 			{ return getS(Field.MCDI); }
        public void setMcdi(String value) 		{ put(Field.MCDI, value); }
        public String getEncoder() 			{ return getS(Field.Encoder); }
        public void setEncoder(String value) 		{ put(Field.Encoder, value); }
        public String getEncodersettings() 		{ return getS(Field.EncoderSettings); }
        public void setEncodersettings(String value) 	{ put(Field.EncoderSettings, value); }
        public String getGroup() 			{ return getS(Field.Group); }
        public void setGroup(String value) 		{ put(Field.Group, value); }
        public String getLanguage() 			{ return getS(Field.Language); }
        public void setLanguage(String value) 		{ put(Field.Language, value); }
        public String getPublisher() 			{ return getS(Field.Publisher); }
        public void setPublisher(String value) 		{ put(Field.Publisher, value); }
        public String getUrl() 			        { return getS(Field.URL); }
        public void setUrl(String value) 		{ put(Field.URL, value); }
        public float getBpm() 		        	{ return getF(Field.BPM); }
        public void setBpm(Float value) 		{ put(Field.BPM, value); }
        public String getUserinfo() 			{ return getS(Field.UserInfo); }
        public void setUserinfo(String value) 		{ put(Field.UserInfo, value); }
        public String getIsrc() 			{ return getS(Field.ISRC); }
        public void setIsrc(String value) 		{ put(Field.ISRC, value); }
        public String getMedia() 			{ return getS(Field.Media); }
        public void setMedia(String value) 		{ put(Field.Media, value); }
        public String getCopyright() 			{ return getS(Field.Copyright); }
        public void setCopyright(String value) 		{ put(Field.Copyright, value); }
        public String getOriginalartist() 		{ return getS(Field.OriginalArtist); }
        public void setOriginalartist(String value) 	{ put(Field.OriginalArtist, value); }
        public String getOriginalalbum() 		{ return getS(Field.OriginalAlbum); }
        public void setOriginalalbum(String value) 	{ put(Field.OriginalAlbum, value); }
        public String getConductor() 			{ return getS(Field.Conductor); }
        public void setConductor(String value) 		{ put(Field.Conductor, value); }
        public String getLyricist() 			{ return getS(Field.Lyricist); }
        public void setLyricist(String value) 		{ put(Field.Lyricist, value); }
        public String getType() 			{ return getS(Field.Type); }
        public String getFile() 			{ return getS(Field.File); }
        public void setFile(String value) 		{ put(Field.File, value); }
        public int getBitrate() 			{ return getI(Field.BitRate); }
        public int getSamplerate() 			{ return getI(Field.SampleRate); }
        public boolean getMono() 			{ return getB(Field.Mono); }
        public boolean getVbr() 			{ return getB(Field.VBR); }
        public int getDuration() 			{ return getI(Field.Duration); }
        public int getFilesize() 			{ return getI(Field.FileSize); }
        public int getLastmodified() 			{ return getI(Field.LastModified); }
        public void setLastmodified(int value) 		{ put(Field.LastModified, value); touched = true; }
        public float getGain() 			        { return getF(Field.Gain); }
        public void setGain(float value) 		{ put(Field.Gain, value); }
        public boolean getGapless() 			{ return getB(Field.Gapless); }
        public void setGapless(boolean value) 		{ put(Field.Gapless, value); }
        public boolean getId3v1() 			{ return getB(Field.ID3v1); }
        public void setId3v1(boolean value) 		{ put(Field.ID3v1, value); }
        public boolean getId3v22() 			{ return getB(Field.ID3v22); }
        public void setId3v22(boolean value) 		{ put(Field.ID3v22, value); }
        public boolean getId3v23() 			{ return getB(Field.ID3v23); }
        public void setId3v23(boolean value) 		{ put(Field.ID3v23, value); }
        public boolean getId3v24() 			{ return getB(Field.ID3v24); }
        public void setId3v24(boolean value) 		{ put(Field.ID3v24, value); }
        public boolean getWarning() 			{ return getB(Field.Warning); }
        public String getImagedescription() 		{ return getS(Field.ImageDescription); }
        public void setImagedescription(String value) 	{ put(Field.ImageDescription, value); }
        public String getImagemimetype() 		{ return getS(Field.ImageMIMEType); }
        public void setImagemimetype(String value) 	{ put(Field.ImageMIMEType, value); }

        public byte[] getImage() throws IOException {
            if (image == null) {
                Number off = map.get(Field.ImageOffset);
                Number len = map.get(Field.ImageLength);
                if (off != null && len != null) {
                    File file = FileUtils.getAbsoluteFile(new File(map.get(Field.File)));
                    FileChannel in = null;
                    try {
                        in = new FileInputStream(file).getChannel();
                        ByteBuffer buf = ByteBuffer.allocate(len.intValue());
                        in.position(off.longValue());
                        in.read(buf);
                        image = buf.array();
                    } finally {
                        in.close();
                    }
                }
            }
            return image;
        }

        private String getS(Field<String> field) {
            String s = map.get(field);
            return s == null ? "" : s;
        }

        private int getI(Field<Integer> field) {
            Integer s = map.get(field);
            return s == null ? 0 : s.intValue();
        }

        private float getF(Field<Float> field) {
            Float s = map.get(field);
            return s == null ? 0 : s.floatValue();
        }

        private boolean getB(Field<Boolean> field) {
            Boolean s = map.get(field);
            return s == Boolean.TRUE;
        }

        private void put(Field<Boolean> field, boolean value) {
            if (value != getB(field)) {
                if (value) {
                    map.put(field, value);
                } else {
                    if (debug > 1) out.println("## Removed "+field.getName()+" value \""+value+"\"");
                    map.remove(field);
                }
                changed = true;
            }
        }

        private void put(Field<String> field, String value) {
            boolean equals = false;
            String old = getS(field);
            if (value == null) {
                value = "";
            }

            if (field == Field.File) {
                if (value.length() != 0 && !value.equals(getS(field))) {
                    try {
                        File oldfile = new File(getS(field));
                        File newfile = new File(value);
                        equals = FileUtils.equals(oldfile, newfile);
                    } catch (IOException e) {
                        throw new IllegalStateException(e);
                    }
                }
            } else {
                equals = old.equals(value);
                changed = !equals;
            }
            if (!equals) {
                if (value.length() == 0) {
                    if (debug > 1) out.println("## Removed "+field.getName()+" was \""+old+"\"");
                    map.remove(field);
                } else if (old.length() == 0) {
                    if (debug > 1) out.println("## Added "+field.getName()+" now \""+value+"\"");
                    map.put(field, value);
                } else {
                    if (debug > 1) out.println("## Changed "+field.getName()+" from \""+old+"\" to \""+value+"\"");
                    map.put(field, value);
                }
            }
        }

        private void put(Field<Integer> field, int value) {
            if (getI(field) != value) {
                if (debug > 1) out.println("## Changed "+field.getName()+" from \""+getI(field)+"\" to \""+value+"\"");
                map.put(field, Integer.valueOf(value));
                changed = true;
            }
        }

        private void put(Field<Float> field, float value) {
            if (getF(field) != value) {
                if (debug > 1) out.println("## Changed "+field.getName()+" from \""+getF(field)+"\" to \""+value+"\"");
                map.put(field, Float.valueOf(value));
                changed = true;
            }
        }

    }

}
